# distutils: libraries = flint
# distutils: depends = flint/qsieve.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/tools/flint-autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    mp_limb_t qsieve_knuth_schroeppel(qs_t qs_inf) noexcept
    mp_limb_t qsieve_primes_init(qs_t qs_inf) noexcept
    mp_limb_t qsieve_primes_increment(qs_t qs_inf, mp_limb_t delta) noexcept
    void qsieve_init_A0(qs_t qs_inf) noexcept
    void qsieve_next_A0(qs_t qs_inf) noexcept
    void qsieve_compute_pre_data(qs_t qs_inf) noexcept
    void qsieve_init_poly_first(qs_t qs_inf) noexcept
    void qsieve_init_poly_next(qs_t qs_inf, slong i) noexcept
    void qsieve_compute_C(fmpz_t C, qs_t qs_inf, qs_poly_t poly) noexcept
    void qsieve_do_sieving(qs_t qs_inf, unsigned char * sieve, qs_poly_t poly) noexcept
    void qsieve_do_sieving2(qs_t qs_inf, unsigned char * sieve, qs_poly_t poly) noexcept
    slong qsieve_evaluate_candidate(qs_t qs_inf, ulong i, unsigned char * sieve, qs_poly_t poly) noexcept
    slong qsieve_evaluate_sieve(qs_t qs_inf, unsigned char * sieve, qs_poly_t poly) noexcept
    slong qsieve_collect_relations(qs_t qs_inf, unsigned char * sieve) noexcept
    void qsieve_write_to_file(qs_t qs_inf, mp_limb_t prime, const fmpz_t Y, const qs_poly_t poly) noexcept
    hash_t * qsieve_get_table_entry(qs_t qs_inf, mp_limb_t prime) noexcept
    void qsieve_add_to_hashtable(qs_t qs_inf, mp_limb_t prime) noexcept
    relation_t qsieve_parse_relation(qs_t qs_inf, char * str) noexcept
    relation_t qsieve_merge_relation(qs_t qs_inf, relation_t  a, relation_t  b) noexcept
    int qsieve_compare_relation(const void * a, const void * b) noexcept
    int qsieve_remove_duplicates(relation_t * rel_list, slong num_relations) noexcept
    void qsieve_insert_relation2(qs_t qs_inf, relation_t * rel_list, slong num_relations) noexcept
    int qsieve_process_relation(qs_t qs_inf) noexcept
    void qsieve_factor(fmpz_factor_t factors, const fmpz_t n) noexcept



from collections.abc import Iterable
from typing import Any

def print_order(lhs: Any, rhs: Any) -> int:
    ...

class _print_key:
    def __init__(self, ex: Any):
        ...

    def __lt__(self, other: '_print_key') -> bool:
        ...

def print_sorted(expressions: Iterable) -> list:
    ...

class _math_key:
    def __init__(self, ex: Any):
        ...

    def __lt__(self, other: '_math_key') -> bool:
        ...

def math_sorted(expressions: Iterable) -> list:
    ...

def mixed_order(lhs: Any, rhs: Any) -> int:
    ...

class _mixed_key:
    def __init__(self, ex: Any):
        ...

    def __lt__(self, other: '_mixed_key') -> bool:
        ...

def mixed_sorted(expressions: Iterable) -> list:
    ...



# sage.doctest: needs sage.combinat sage.modules
r"""
Bijection classes for type `D_4^{(3)}`

Part of the (internal) classes which runs the bijection between rigged
configurations and KR tableaux of type `D_4^{(3)}`.

AUTHORS:

- Travis Scrimshaw (2014-09-10): Initial version

TESTS::

    sage: KRT = crystals.TensorProductOfKirillovReshetikhinTableaux(['D', 4, 3], [[2, 1]])
    sage: from sage.combinat.rigged_configurations.bij_type_D_tri import KRTToRCBijectionTypeDTri
    sage: bijection = KRTToRCBijectionTypeDTri(KRT(pathlist=[[-1,2]]))
    sage: TestSuite(bijection).run()
    sage: RC = RiggedConfigurations(['D', 4, 3], [[2, 1]])
    sage: from sage.combinat.rigged_configurations.bij_type_D_tri import RCToKRTBijectionTypeDTri
    sage: bijection = RCToKRTBijectionTypeDTri(RC(partition_list=[[],[]]))
    sage: TestSuite(bijection).run()
"""

# ****************************************************************************
#       Copyright (C) 2014 Travis Scrimshaw <tscrim@ucdavis.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.combinat.rigged_configurations.bij_type_A import KRTToRCBijectionTypeA
from sage.combinat.rigged_configurations.bij_type_A import RCToKRTBijectionTypeA


class KRTToRCBijectionTypeDTri(KRTToRCBijectionTypeA):
    r"""
    Specific implementation of the bijection from KR tableaux to rigged
    configurations for type `D_4^{(3)}`.

    This inherits from type `A_n^{(1)}` because we use the same methods in
    some places.
    """

    def next_state(self, val):
        r"""
        Build the next state for type `D_4^{(3)}`.

        TESTS::

            sage: KRT = crystals.TensorProductOfKirillovReshetikhinTableaux(['D', 4, 3], [[2,1]])
            sage: from sage.combinat.rigged_configurations.bij_type_D_tri import KRTToRCBijectionTypeDTri
            sage: bijection = KRTToRCBijectionTypeDTri(KRT(pathlist=[[-1,2]]))
            sage: bijection.cur_path.insert(0, [])
            sage: bijection.cur_dims.insert(0, [0, 1])
            sage: bijection.cur_path[0].insert(0, [2])
            sage: bijection.next_state(2)
        """
        tableau_height = len(self.cur_path[0]) - 1

        if val == 'E':
            self.ret_rig_con[0].insert_cell(0)
            self.ret_rig_con[1].insert_cell(0)
            if tableau_height == 0:
                self.ret_rig_con[0].insert_cell(0)
            self._update_vacancy_nums(0)
            self._update_vacancy_nums(1)
            self._update_partition_values(0)
            self._update_partition_values(1)
            return

        if val > 0:
            # If it is a regular value, we follow the A_n rules
            KRTToRCBijectionTypeA.next_state(self, val)
            return

        pos_val = -val

        if pos_val == 0:
            if len(self.ret_rig_con[0]) > 0:
                max_width = self.ret_rig_con[0][0]
            else:
                max_width = 1
            max_width = self.ret_rig_con[0].insert_cell(max_width)
            width_n = max_width + 1

            # Follow regular A_n rules
            for a in reversed(range(tableau_height, 2)):
                max_width = self.ret_rig_con[a].insert_cell(max_width)
            self._update_vacancy_nums(0)
            self._update_partition_values(0)
            self._update_vacancy_nums(1)
            self._update_partition_values(1)

            # Make the largest string at \nu^{(1)} quasi-singular
            p = self.ret_rig_con[0]
            num_rows = len(p)
            for i in range(num_rows):
                if p._list[i] == width_n:
                    j = i+1
                    while j < num_rows and p._list[j] == width_n \
                      and p.vacancy_numbers[j] == p.rigging[j]:
                        j += 1
                    p.rigging[j-1] -= 1
                    break
            return

        case_S = [None] * 2
        pos_val = -val

        if pos_val < 3:
            # Always add a cell to the first singular value in the first
            #   tableau we are updating.
            if len(self.ret_rig_con[pos_val - 1]) > 0:
                max_width = self.ret_rig_con[pos_val - 1][0]
            else:
                max_width = 1

            # Add cells similar to type A_n but we move to the right
            for a in range(pos_val - 1, 2):
                max_width = self.ret_rig_con[a].insert_cell(max_width)
                case_S[a] = max_width
        else:
            if len(self.ret_rig_con[0]) > 0:
                max_width = self.ret_rig_con[0][0]
            else:
                max_width = 1

        # Special case for going through 0
        # If we find a quasi-singular string first, then we are in case (Q, S)
        #   otherwise we will find a singular string and insert 2 cells
        P = self.ret_rig_con[0]
        num_rows = len(P)
        case_QS = False
        for i in range(num_rows + 1):
            if i == num_rows:
                max_width = 0
                if case_QS:
                    P._list.append(1)
                    P.vacancy_numbers.append(None)
                    # Go through our partition until we find a length of greater than 1
                    j = len(P._list) - 1
                    while j >= 0 and P._list[j] == 1:
                        j -= 1
                    P.rigging.insert(j + 1, None)
                    width_n = 1
                else:
                    # Go through our partition until we find a length of greater than 2
                    j = len(P._list) - 1
                    while j >= 0 and P._list[j] <= 2:
                        j -= 1
                    P._list.insert(j+1, 2)
                    P.vacancy_numbers.insert(j+1, None)
                    P.rigging.insert(j+1, None)
                break
            elif P._list[i] <= max_width:
                if P.vacancy_numbers[i] == P.rigging[i]:
                    max_width = P._list[i]
                    if case_QS:
                        P._list[i] += 1
                        width_n = P._list[i]
                        P.rigging[i] = None
                    else:
                        j = i - 1
                        while j >= 0 and P._list[j] <= max_width + 2:
                            P.rigging[j+1] = P.rigging[j]  # Shuffle it along
                            j -= 1
                        P._list.pop(i)
                        P._list.insert(j+1, max_width + 2)
                        P.rigging[j+1] = None
                    break
                elif P.vacancy_numbers[i] - 1 == P.rigging[i] and not case_QS:
                    case_QS = True
                    P._list[i] += 1
                    P.rigging[i] = None
                    # No need to set max_width here since we will find a singular string

        # Now go back following the regular C_n (ish) rules
        if case_S[1] == max_width:
            P = self.ret_rig_con[1]

            # Special case when adding twice to the first row
            if P.rigging[0] is None:
                P._list[0] += 1
            else:
                for i in reversed(range(1, len(P))):
                    if P.rigging[i] is None:
                        j = i - 1
                        while j >= 0 and P._list[j] == P._list[i]:
                            P.rigging[j+1] = P.rigging[j]  # Shuffle it along
                            j -= 1
                        P._list[j+1] += 1
                        P.rigging[j+1] = None
                        break
        else:
            max_width = self.ret_rig_con[1].insert_cell(max_width)

        if tableau_height == 0:
            if case_S[0] == max_width:
                P = self.ret_rig_con[0]
                # Since this is on the way back, the added string we want
                #   to bump will never be the first (largest) string
                for i in reversed(range(1, len(P))):
                    if P.rigging[i] is None:
                        j = i - 1
                        while j >= 0 and P._list[j] == P._list[i]:
                            P.rigging[j+1] = P.rigging[j]  # Shuffle it along
                            j -= 1
                        P._list[j+1] += 1
                        P.rigging[j+1] = None
                        break
            else:
                max_width = self.ret_rig_con[0].insert_cell(max_width)

        self._update_vacancy_nums(0)
        self._update_partition_values(0)
        self._update_vacancy_nums(1)
        self._update_partition_values(1)

        if case_QS:
            # Make the new string quasi-singular
            num_rows = len(P)
            for i in range(num_rows):
                if P._list[i] == width_n:
                    j = i+1
                    while j < num_rows and P._list[j] == width_n \
                      and P.vacancy_numbers[j] == P.rigging[j]:
                        j += 1
                    P.rigging[j-1] -= 1
                    break


class RCToKRTBijectionTypeDTri(RCToKRTBijectionTypeA):
    r"""
    Specific implementation of the bijection from rigged configurations to
    tensor products of KR tableaux for type `D_4^{(3)}`.
    """

    def next_state(self, height):
        r"""
        Build the next state for type `D_4^{(3)}`.

        TESTS::

            sage: RC = RiggedConfigurations(['D', 4, 3], [[2, 1]])
            sage: from sage.combinat.rigged_configurations.bij_type_D_tri import RCToKRTBijectionTypeDTri
            sage: bijection = RCToKRTBijectionTypeDTri(RC(partition_list=[[3],[2]]))
            sage: bijection.next_state(2)
            -3
        """
        height -= 1  # indexing
        ell = [None] * 6
        case_S = [False] * 3
        case_Q = False
        b = None

        # Calculate the rank and ell values

        last_size = 0
        for a in range(height, 2):
            ell[a] = self._find_singular_string(self.cur_partitions[a], last_size)

            if ell[a] is None:
                b = a + 1
                break
            else:
                last_size = self.cur_partitions[a][ell[a]]

        if b is None:
            partition = self.cur_partitions[0]
            # Modified version of _find_singular_string()
            for i in reversed(range(len(partition))):
                if partition[i] >= last_size:
                    if partition.vacancy_numbers[i] == partition.rigging[i] and i != ell[0]:
                        if partition[i] == 1:
                            b = 'E'
                        else:
                            last_size = partition[i]
                        case_S[2] = True
                        ell[3] = i
                        break
                    elif partition.vacancy_numbers[i] - 1 == partition.rigging[i] and not case_Q:
                        case_Q = True
                        # Check if the block is singular
                        block_size = partition[i]
                        for j in reversed(range(i)):
                            if partition[j] != block_size:
                                break
                            elif partition.vacancy_numbers[j] == partition.rigging[j] and j != ell[0]:
                                case_Q = False
                                break
                        if case_Q:
                            last_size = partition[i] + 1
                            ell[2] = i

            if ell[3] is None:
                if not case_Q:
                    b = 3
                else:
                    b = 0

        if b is None:  # Going back
            if self.cur_partitions[1][ell[1]] == last_size:
                ell[4] = ell[1]
                case_S[1] = True
            else:
                ell[4] = self._find_singular_string(self.cur_partitions[1], last_size)

                if ell[4] is None:
                    b = -3
                else:
                    last_size = self.cur_partitions[1][ell[4]]

        if b is None:  # Final partition
            P = self.cur_partitions[0]
            if ell[0] is not None and P[ell[0]] == last_size:
                ell[5] = ell[0]
                case_S[0] = True
            else:
                # Modified form of _find_singular_string
                end = ell[3]
                for i in reversed(range(end)):
                    if P[i] >= last_size and P.vacancy_numbers[i] == P.rigging[i]:
                        ell[5] = i
                        break

                if ell[5] is None:
                    b = -2

        if b is None:
            b = -1

        # Determine the new rigged configuration by removing boxes from the
        #   selected string and then making the new string singular
        if case_S[1]:
            row1 = [self.cur_partitions[1].remove_cell(ell[4], 2)]
        else:
            row1 = [self.cur_partitions[1].remove_cell(ell[1]),
                    self.cur_partitions[1].remove_cell(ell[4])]

        if case_S[0]:
            row0 = [self.cur_partitions[0].remove_cell(ell[5], 2)]
            row0.append(self.cur_partitions[0].remove_cell(ell[3], 2))
        else:
            if case_Q:
                if ell[0] is None or ell[0] < ell[2]:
                    row0 = [self.cur_partitions[0].remove_cell(ell[2]),
                            self.cur_partitions[0].remove_cell(ell[0])]
                else:
                    row0 = [self.cur_partitions[0].remove_cell(ell[0]),
                            self.cur_partitions[0].remove_cell(ell[2])]
                if case_S[2]:
                    quasi = self.cur_partitions[0].remove_cell(ell[3])
            else:
                row0 = [self.cur_partitions[0].remove_cell(ell[0])]
                if case_S[2]:
                    row0.append(self.cur_partitions[0].remove_cell(ell[3], 2))

            row0.append(self.cur_partitions[0].remove_cell(ell[5]))

        self._update_vacancy_numbers(0)
        self._update_vacancy_numbers(1)

        for l in row1:
            if l is not None:
                self.cur_partitions[1].rigging[l] = self.cur_partitions[1].vacancy_numbers[l]
        for l in row0:
            if l is not None:
                self.cur_partitions[0].rigging[l] = self.cur_partitions[0].vacancy_numbers[l]

        # If case (Q,S) holds, then we must make the larger string quasisingular
        if case_Q and case_S[2]:
            P = self.cur_partitions[0]
            vac_num = P.vacancy_numbers[quasi]
            P.rigging[quasi] = vac_num
            block_len = P[quasi]
            j = quasi + 1
            length = len(P)
            # Find the place for the quasisingular rigging
            while j < length and P[j] == block_len and P.rigging[j] == vac_num:
                j += 1
            P.rigging[j-1] = vac_num - 1

        return b



from sage.libs.mpfi.types cimport mpfi_ptr

cdef int mpfi_set_sage(mpfi_ptr re, mpfi_ptr im, x, field, int base) except -1
cdef int mpfi_interv_sage(mpfi_ptr re, mpfi_ptr im, x, y, field, int base) except -1
cdef int mpfi_set_via_RR(mpfi_ptr re, x, field) except -1



"""
TESTS::

    sage: R = Zp(5, prec=5, type='fixed-mod')
    sage: a = random_matrix(R,5)                                                        # needs sage.geometry.polyhedron
    sage: a.determinant().parent() is R                                                 # needs sage.geometry.polyhedron
    True
    sage: K = Qp(3, 10,'capped-rel'); K.krull_dimension()
    0

    Computation of logs::

    sage: Qp(5)(1).log()
    O(5^20)
    sage: Qp(5)(-1).log()
    O(5^20)
    sage: Qp(5,prec=5)(7).log()
    4*5^2 + 4*5^3 + 3*5^4 + O(5^5)
    sage: Qp(5,prec=10)(25*8).log(p_branch = 0)
    5 + 3*5^3 + 3*5^4 + 2*5^5 + 5^6 + 2*5^7 + 3*5^8 + 3*5^9 + O(5^10)
    sage: Zp(5,prec=10)(8).log()
    5 + 3*5^3 + 3*5^4 + 2*5^5 + 5^6 + 2*5^7 + 3*5^8 + 3*5^9 + O(5^10)

    Loading and saving elements of various types::

    sage: a = Zp(5)(-3); loads(dumps(a)) == a
    True

    sage: M = MatrixSpace(pAdicField(3,100),2)                                          # needs sage.geometry.polyhedron
    sage: (M([1,0,0,90]) - (1+O(3^100)) * M(1)).left_kernel()                           # needs sage.geometry.polyhedron
    Vector space of degree 2 and dimension 1 over 3-adic Field with capped relative precision 100
    Basis matrix:
    [1 + O(3^100)            0]
"""



"""
Optimized computing of relation matrices in certain cases
"""

#############################################################################
#       Copyright (C) 2010 William Stein <wstein@gmail.com>
#  Distributed under the terms of the GNU General Public License (GPL) v2+.
#  The full text of the GPL is available at:
#                  https://www.gnu.org/licenses/
#############################################################################

from sage.misc.verbose import verbose
from sage.rings.rational cimport Rational


def sparse_2term_quotient_only_pm1(rels, n):
    r"""
    Perform Sparse Gauss elimination on a matrix all of whose columns
    have at most 2 nonzero entries with relations all 1 or -1.

    This algorithm is more subtle than just "identify symbols in pairs",
    since complicated relations can cause generators to equal 0.

    .. NOTE::

        Note the condition on the s,t coefficients in the relations
        being 1 or -1 for this optimized function.  There is a more
        general function in relation_matrix.py, which is much, much
        slower.

    INPUT:

    - ``rels`` -- iterable made of pairs ((i,s), (j,t)). The pair
      represents the relation `s x_i + t x_j = 0`, where the `i, j` must
      be Python int's, and the `s, t` must all be 1 or -1.

    - ``n`` -- integer; the `x_i` are `x_0, \ldots, x_{n-1}`

    OUTPUT:

    - ``mod`` -- list such that mod[i] = (j,s), which means that x_i
      is equivalent to s*x_j, where the x_j are a basis for the
      quotient.

    The output depends on the order of the input.

    EXAMPLES::

        sage: from sage.modular.modsym.relation_matrix_pyx import sparse_2term_quotient_only_pm1
        sage: rels = [((0,1), (1,-1)), ((1,1), (3,1)), ((2,1),(3,1)), ((4,1),(5,-1))]
        sage: n = 6
        sage: sparse_2term_quotient_only_pm1(rels, n)
        [(3, -1), (3, -1), (3, -1), (3, 1), (5, 1), (5, 1)]
    """
    n = int(n)

    tm = verbose("Starting optimized integer sparse 2-term quotient...")

    cdef int c0, c1, i, die
    cdef list free = list(range(n))
    cdef list coef = [1] * n
    cdef list related_to_me = [[] for i in range(n)]

    for v0, v1 in rels:
        c0 = coef[v0[0]] * v0[1]
        c1 = coef[v1[0]] * v1[1]

        # Mod out by the following relation:
        #
        #    c0*free[v0[0]] + c1*free[v1[0]] = 0.
        #
        die = -1
        if c0 == 0 and c1 == 0:
            pass
        elif c0 == 0 and c1 != 0:  # free[v1[0]] --> 0
            die = free[v1[0]]
        elif c1 == 0 and c0 != 0:
            die = free[v0[0]]
        elif free[v0[0]] == free[v1[0]]:
            if c0 + c1 != 0:
                # all xi equal to free[v0[0]] must now equal to zero.
                die = free[v0[0]]
        else:  # x1 = -c1/c0 * x2.
            x = free[v0[0]]
            free[x] = free[v1[0]]
            if c0 != 1 and c0 != -1:
                raise ValueError("coefficients must all be -1 or 1.")
            coef[x] = -c1 * c0
            for i in related_to_me[x]:
                free[i] = free[x]
                coef[i] *= coef[x]
                related_to_me[free[v1[0]]].append(i)
            related_to_me[free[v1[0]]].append(x)
        if die != -1:
            for i in related_to_me[die]:
                free[i] = 0
                coef[i] = 0
            free[die] = 0
            coef[die] = 0

    # Special casing the rationals leads to a huge speedup,
    # actually.  (All the code above is slower than just this line
    # without this special case.)
    mod = [(fi, Rational(ci)) for fi, ci in zip(free, coef)]

    verbose("finished", tm)
    return mod



# sage.doctest: needs sage.symbolic
r"""
C-finite sequences

C-finite infinite sequences satisfy homogeneous linear recurrences with constant coefficients:

.. MATH::

    a_{n+d} = c_0a_n + c_1a_{n+1} + \cdots + c_{d-1}a_{n+d-1}, \quad d>0.

They are completely defined by their ordinary generating function (o.g.f., which
is always a :mod:`fraction <sage.rings.fraction_field_element>` of
:mod:`polynomials <sage.rings.polynomial.polynomial_element>` over `\ZZ` or `\QQ` ).

EXAMPLES::

    sage: fibo = CFiniteSequence(x/(1-x-x^2))        # the Fibonacci sequence
    sage: fibo
    C-finite sequence, generated by -x/(x^2 + x - 1)
    sage: fibo.parent()
    The ring of C-Finite sequences in x over Rational Field
    sage: fibo.parent().category()
    Category of commutative rings
    sage: C.<x> = CFiniteSequences(QQ)
    sage: fibo.parent() == C
    True
    sage: C
    The ring of C-Finite sequences in x over Rational Field
    sage: C(x/(1-x-x^2))
    C-finite sequence, generated by -x/(x^2 + x - 1)
    sage: C(x/(1-x-x^2)) == fibo
    True
    sage: var('y')
    y
    sage: CFiniteSequence(y/(1-y-y^2))
    C-finite sequence, generated by -y/(y^2 + y - 1)
    sage: CFiniteSequence(y/(1-y-y^2)) == fibo
    False

Finite subsets of the sequence are accessible via python slices::

    sage: fibo[137]                 #the 137th term of the Fibonacci sequence
    19134702400093278081449423917
    sage: fibo[137] == fibonacci(137)
    True
    sage: fibo[0:12]
    [0, 1, 1, 2, 3, 5, 8, 13, 21, 34, 55, 89]
    sage: fibo[14:4:-2]
    [377, 144, 55, 21, 8]

They can be created also from the coefficients and start values of a recurrence::

    sage: r = C.from_recurrence([1,1],[0,1])
    sage: r == fibo
    True

Given enough values, the o.g.f. of a C-finite sequence
can be guessed::

    sage: r = C.guess([0,1,1,2,3,5,8])
    sage: r == fibo
    True

.. SEEALSO::

    :func:`fibonacci`, :class:`BinaryRecurrenceSequence`

.. TODO::

    Implement a feature enabling::

        sage: # not implemented
        sage: CFiniteSequence(x+x^2+x^3+x^4+x^5+O(x^6))
        sage: latex(r)
        \big\{a_{n\ge0}\big|a_{n+2}=\sum_{i=0}^{1}c_ia_{n+i}, c=\{1,1\}, a_{n<2}=\{0,0,0,1\}\big\}
        sage: r.egf()
        exp(2*x)
        sage: r = CFiniteSequence(1/(1-y-x*y), x)

REFERENCES:

- [GK1982]_
- [KP2011]_
- [SZ1994]_
- [Zei2011]_

AUTHORS:

- Ralf Stephan (2014): initial version
"""

# ****************************************************************************
#       Copyright (C) 2014 Ralf Stephan <gtrwst9@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from numbers import Integral

from sage.categories.rings import Rings
from sage.libs.pari import pari
from sage.misc.inherit_comparison import InheritComparisonClasscallMetaclass
from sage.rings.integer_ring import ZZ
from sage.rings.rational_field import QQ
from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
from sage.rings.polynomial.polynomial_ring import PolynomialRing_generic
from sage.rings.laurent_series_ring import LaurentSeriesRing
from sage.rings.power_series_ring import PowerSeriesRing
from sage.rings.fraction_field import FractionField
from sage.structure.element import FieldElement, parent
from sage.structure.parent import Parent
from sage.structure.unique_representation import UniqueRepresentation


def CFiniteSequences(base_ring, names=None, category=None):
    r"""
    Return the commutative ring of C-Finite sequences.

    The ring is defined over a base ring (`\ZZ` or `\QQ` )
    and each element is represented by its ordinary generating function (ogf)
    which is a rational function over the base ring.

    INPUT:

    - ``base_ring`` -- the base ring to construct the fraction field
      representing the C-Finite sequences
    - ``names`` -- (optional) the list of variables

    EXAMPLES::

        sage: C.<x> = CFiniteSequences(QQ)
        sage: C
        The ring of C-Finite sequences in x over Rational Field
        sage: C.an_element()
        C-finite sequence, generated by (x - 2)/(x^2 + x - 1)
        sage: C.category()
        Category of commutative rings
        sage: C.one()
        Finite sequence [1], offset = 0
        sage: C.zero()
        Constant infinite sequence 0.
        sage: C(x)
        Finite sequence [1], offset = 1
        sage: C(1/x)
        Finite sequence [1], offset = -1
        sage: C((-x + 2)/(-x^2 - x + 1))
        C-finite sequence, generated by (x - 2)/(x^2 + x - 1)

    TESTS::

        sage: TestSuite(C).run()
    """
    if isinstance(base_ring, PolynomialRing_generic):
        polynomial_ring = base_ring
        base_ring = polynomial_ring.base_ring()
    if names is None:
        names = ['x']
    elif len(names) > 1:
        raise NotImplementedError("Multidimensional o.g.f. not implemented.")
    if category is None:
        category = Rings().Commutative()
    if base_ring not in [QQ, ZZ]:
        raise ValueError("O.g.f. base not rational.")
    polynomial_ring = PolynomialRing(base_ring, names)
    return CFiniteSequences_generic(polynomial_ring, category)


class CFiniteSequence(FieldElement,
                      metaclass=InheritComparisonClasscallMetaclass):
    r"""
    Create a C-finite sequence given its ordinary generating function.

    INPUT:

    - ``ogf`` -- a rational function, the ordinary generating function
      (can be an element from the symbolic ring, fraction field or polynomial
      ring)

    OUTPUT: a CFiniteSequence object

    EXAMPLES::

        sage: CFiniteSequence((2-x)/(1-x-x^2))     # the Lucas sequence
        C-finite sequence, generated by (x - 2)/(x^2 + x - 1)
        sage: CFiniteSequence(x/(1-x)^3)           # triangular numbers
        C-finite sequence, generated by -x/(x^3 - 3*x^2 + 3*x - 1)

    Polynomials are interpreted as finite sequences, or recurrences of degree 0::

        sage: CFiniteSequence(x^2-4*x^5)
        Finite sequence [1, 0, 0, -4], offset = 2
        sage: CFiniteSequence(1)
        Finite sequence [1], offset = 0

    This implementation allows any polynomial fraction as o.g.f. by interpreting
    any power of `x` dividing the o.g.f. numerator or denominator as a right or left shift
    of the sequence offset::

        sage: CFiniteSequence(x^2+3/x)
        Finite sequence [3, 0, 0, 1], offset = -1
        sage: CFiniteSequence(1/x+4/x^3)
        Finite sequence [4, 0, 1], offset = -3
        sage: P = LaurentPolynomialRing(QQ.fraction_field(), 'X')
        sage: X=P.gen()
        sage: CFiniteSequence(1/(1-X))
        C-finite sequence, generated by -1/(X - 1)

    The o.g.f. is always normalized to get a denominator constant coefficient of `+1`::

        sage: CFiniteSequence(1/(x-2))
        C-finite sequence, generated by 1/(x - 2)

    The given ``ogf`` is used to create an appropriate parent: it can
    be a symbolic expression, a polynomial , or a fraction field element
    as long as it can be coerced into a proper fraction field over the
    rationals::

        sage: var('x')
        x
        sage: f1 = CFiniteSequence((2-x)/(1-x-x^2))
        sage: P.<x> = QQ[]
        sage: f2 = CFiniteSequence((2-x)/(1-x-x^2))
        sage: f1 == f2
        True
        sage: f1.parent()
        The ring of C-Finite sequences in x over Rational Field
        sage: f1.ogf().parent()
        Fraction Field of Univariate Polynomial Ring in x over Rational Field
        sage: CFiniteSequence(log(x))
        Traceback (most recent call last):
        ...
        TypeError: unable to convert log(x) to a rational

    TESTS::

        sage: P.<x> = QQ[]
        sage: CFiniteSequence(0.1/(1-x))
        C-finite sequence, generated by -1/10/(x - 1)
        sage: CFiniteSequence(pi/(1-x))
        Traceback (most recent call last):
        ...
        TypeError: unable to convert -pi to a rational
        sage: P.<x,y> = QQ[]
        sage: CFiniteSequence(x*y)
        Traceback (most recent call last):
        ...
        NotImplementedError: Multidimensional o.g.f. not implemented.
    """
    @staticmethod
    def __classcall_private__(cls, ogf):
        r"""
        Ensure that elements created by :class:`CFiniteSequence` have the same
        parent than the ones created by the parent itself and follow the category
        framework (they should be instance of :class:`CFiniteSequences` automatic
        element class).

        This method is called before the ``__init__`` method, it checks the
        o.g.f to create the appropriate parent.

        INPUT:

        - ``ogf`` -- a rational function

        TESTS::

            sage: f1 = CFiniteSequence((2-x)/(1-x-x^2))
            sage: f1
            C-finite sequence, generated by (x - 2)/(x^2 + x - 1)
            sage: C.<x> = CFiniteSequences(QQ)
            sage: f2 = CFiniteSequence((2-x)/(1-x-x^2))
            sage: f2
            C-finite sequence, generated by (x - 2)/(x^2 + x - 1)
            sage: f3 = C((2-x)/(1-x-x^2))
            sage: f3
            C-finite sequence, generated by (x - 2)/(x^2 + x - 1)
            sage: f1 == f2 and f2 == f3
            True
            sage: f1.parent() == f2.parent() and f2.parent() == f3.parent()
            True
            sage: type(f1)
            <class 'sage.rings.cfinite_sequence.CFiniteSequences_generic_with_category.element_class'>
            sage: type(f1) == type(f2) and type(f2) == type(f3)
            True
            sage: CFiniteSequence(log(x))
            Traceback (most recent call last):
            ...
            TypeError: unable to convert log(x) to a rational
            sage: CFiniteSequence(pi)
            Traceback (most recent call last):
            ...
            TypeError: unable to convert pi to a rational
            sage: var('y')
            y
            sage: f4 = CFiniteSequence((2-y)/(1-y-y^2))
            sage: f4
            C-finite sequence, generated by (y - 2)/(y^2 + y - 1)
            sage: f4 == f1
            False
            sage: f4.parent() == f1.parent()
            False
            sage: f4.parent()
            The ring of C-Finite sequences in y over Rational Field
        """
        br = ogf.base_ring()
        if br not in [QQ, ZZ]:
            br = QQ  # if the base ring of the o.g.f is not QQ, we force it to QQ and see if the o.g.f converts nicely

        # trying to figure out the ogf variables
        variables = []
        if ogf not in br:
            if hasattr(ogf, 'variables'):
                variables = ogf.variables()
            elif hasattr(ogf.parent(), 'gens'):
                variables = ogf.parent().gens()
                # for some reason, fraction field elements don't have the variables
                # method, but symbolic elements don't have the gens method so we check both

        if not variables:
            parent = CFiniteSequences(QQ)  # if we cannot find variables, we create the default parent (with x)
        else:
            parent = CFiniteSequences(QQ, variables)
        return parent(ogf)  # if ogf cannot be converted to a fraction field, this will break and raise the proper error

    def __init__(self, parent, ogf):
        r"""
        Initialize the C-Finite sequence.

        The ``__init__`` method can only be called by the :class:`CFiniteSequences`
        class. By Default, a class call reaches the ``__classcall_private__``
        which first creates a proper parent and then call the ``__init__``.

        INPUT:

        - ``ogf`` -- the ordinary generating function, a fraction of
          polynomials over the rationals
        - ``parent`` -- the parent of the C-Finite sequence, an occurrence of
          :class:`CFiniteSequences`

        OUTPUT: a CFiniteSequence object

        TESTS::

            sage: C.<x> = CFiniteSequences(QQ)
            sage: C((2-x)/(1-x-x^2))            # indirect doctest
            C-finite sequence, generated by (x - 2)/(x^2 + x - 1)
        """

        br = parent.base_ring()
        ogf = parent.fraction_field()(ogf)
        P = parent.polynomial_ring()
        num = ogf.numerator()
        den = ogf.denominator()

        FieldElement.__init__(self, parent)

        if den == 1:
            self._c = []
            self._off = num.valuation()
            self._deg = 0
            if ogf == 0:
                self._a = [0]
            else:
                self._a = num.shift(-self._off).list()
        else:
            # Transform the ogf numerator and denominator to canonical form
            # to get the correct offset, degree, and recurrence coeffs and
            # start values.
            self._off = 0
            self._deg = 0
            if num.constant_coefficient() == 0:
                self._off = num.valuation()
                num = num.shift(-self._off)
            elif den.constant_coefficient() == 0:
                self._off = -den.valuation()
                den = den.shift(self._off)
            f = den.constant_coefficient()
            num = P(num / f)
            den = P(den / f)
            f = num.gcd(den)
            num = P(num / f)
            den = P(den / f)
            self._deg = den.degree()
            self._c = [-den[i] for i in range(1, self._deg + 1)]
            if self._off >= 0:
                num = num.shift(self._off)
            else:
                den = den.shift(-self._off)

            # determine start values (may be different from _get_item_ values)
            alen = max(self._deg, num.degree() + 1)
            R = LaurentSeriesRing(br, parent.variable_name(), default_prec=alen)
            if den != 1:
                self._a = R(num / den).list()
            else:
                self._a = num.list()
            if len(self._a) < alen:
                self._a.extend([0] * (alen - len(self._a)))

            ogf = num / den

        self._ogf = ogf

    def _repr_(self) -> str:
        """
        Return textual definition of sequence.

        TESTS::

            sage: CFiniteSequence(1/x^5)
            Finite sequence [1], offset = -5
            sage: CFiniteSequence(x^3)
            Finite sequence [1], offset = 3
        """
        if self._deg == 0:
            if self.ogf() == 0:
                return 'Constant infinite sequence 0.'
            return 'Finite sequence ' + str(self._a) + ', offset = ' + str(self._off)
        return 'C-finite sequence, generated by ' + str(self.ogf())

    def __hash__(self):
        r"""
        Hash value for C finite sequence.

        EXAMPLES::

            sage: C.<x> = CFiniteSequences(QQ)
            sage: hash(C((2-x)/(1-x-x^2)))  # random
            42
        """
        return hash(self.parent()) ^ hash(self._ogf)

    def _add_(self, other):
        """
        Addition of C-finite sequences.

        TESTS::

            sage: C.<x> = CFiniteSequences(QQ)
            sage: r = C(1/(1-2*x))
            sage: r[0:5]                  # a(n) = 2^n
            [1, 2, 4, 8, 16]
            sage: s = C.from_recurrence([1],[1])
            sage: (r + s)[0:5]            # a(n) = 2^n + 1
            [2, 3, 5, 9, 17]
            sage: r + 0 == r
            True
            sage: (r + x^2)[0:5]
            [1, 2, 5, 8, 16]
            sage: (r + 3/x)[-1]
            3
            sage: r = CFiniteSequence(x)
            sage: r + 0 == r
            True
            sage: CFiniteSequence(0) + CFiniteSequence(0)
            Constant infinite sequence 0.
        """
        return CFiniteSequence(self.ogf() + other.numerator() / other.denominator())

    def _sub_(self, other):
        """
        Subtraction of C-finite sequences.

        TESTS::

            sage: C.<x> = CFiniteSequences(QQ)
            sage: r = C(1/(1-2*x))
            sage: r[0:5]                  # a(n) = 2^n
            [1, 2, 4, 8, 16]
            sage: s = C.from_recurrence([1],[1])
            sage: (r - s)[0:5]            # a(n) = 2^n + 1
            [0, 1, 3, 7, 15]
        """
        return CFiniteSequence(self.ogf() - other.numerator() / other.denominator())

    def _mul_(self, other):
        """
        Multiplication of C-finite sequences.

        TESTS::

            sage: C.<x> = CFiniteSequences(QQ)
            sage: r = C.guess([1,2,3,4,5,6])
            sage: (r*r)[0:6]                  # self-convolution
            [1, 4, 10, 20, 35, 56]
            sage: r = C(x)
            sage: r*1 == r
            True
            sage: r*-1
            Finite sequence [-1], offset = 1
            sage: C(0) * C(1)
            Constant infinite sequence 0.
        """
        return CFiniteSequence(self.ogf() * other.numerator() / other.denominator())

    def _div_(self, other):
        """
        Division of C-finite sequences.

        TESTS::

            sage: C.<x> = CFiniteSequences(QQ)
            sage: r = C.guess([1,2,3,4,5,6])
            sage: (r/2)[0:6]
            [1/2, 1, 3/2, 2, 5/2, 3]
            sage: s = C(x)
            sage: s/(s*-1 + 1)
            C-finite sequence, generated by -x/(x - 1)
        """
        return CFiniteSequence(self.ogf() / (other.numerator() / other.denominator()))

    def coefficients(self):
        """
        Return the coefficients of the recurrence representation of the
        C-finite sequence.

        OUTPUT: list of values

        EXAMPLES::

            sage: C.<x> = CFiniteSequences(QQ)
            sage: lucas = C((2-x)/(1-x-x^2))   # the Lucas sequence
            sage: lucas.coefficients()
            [1, 1]
        """
        return self._c

    def __eq__(self, other):
        """
        Compare two CFiniteSequences.

        EXAMPLES::

            sage: f = CFiniteSequence((2-x)/(1-x-x^2))
            sage: f2 = CFiniteSequence((2-x)/(1-x-x^2))
            sage: f == f2
            True
            sage: f == (2-x)/(1-x-x^2)
            False
            sage: (2-x)/(1-x-x^2) == f
            False
            sage: C.<x> = CFiniteSequences(QQ)
            sage: r = C.from_recurrence([1,1],[2,1])
            sage: s = C.from_recurrence([-1],[1])
            sage: r == s
            False
            sage: r = C.from_recurrence([-1],[1])
            sage: s = C(1/(1+x))
            sage: r == s
            True
        """
        if not isinstance(other, CFiniteSequence):
            return False
        return self.ogf() == other.ogf()

    def __ne__(self, other):
        """
        Compare two CFiniteSequences.

        EXAMPLES::

            sage: f = CFiniteSequence((2-x)/(1-x-x^2))
            sage: f2 = CFiniteSequence((2-x)/(1-x-x^2))
            sage: f != f2
            False
            sage: f != (2-x)/(1-x-x^2)
            True
            sage: (2-x)/(1-x-x^2) != f
            True
            sage: C.<x> = CFiniteSequences(QQ)
            sage: r = C.from_recurrence([1,1],[2,1])
            sage: s = C.from_recurrence([-1],[1])
            sage: r != s
            True
            sage: r = C.from_recurrence([-1],[1])
            sage: s = C(1/(1+x))
            sage: r != s
            False
        """
        return not self.__eq__(other)

    def __getitem__(self, key):
        r"""
        Return a slice of the sequence.

        EXAMPLES::

            sage: C.<x> = CFiniteSequences(QQ)
            sage: r = C.from_recurrence([3,3],[2,1])
            sage: r[2]
            9
            sage: r[101]
            16158686318788579168659644539538474790082623100896663971001
            sage: r = C(1/(1-x))
            sage: r[5]
            1
            sage: r = C(x)
            sage: r[0]
            0
            sage: r[1]
            1
            sage: r = C(0)
            sage: r[66]
            0
            sage: lucas = C.from_recurrence([1,1],[2,1])
            sage: lucas[5:10]
            [11, 18, 29, 47, 76]
            sage: r = C((2-x)/x/(1-x-x*x))
            sage: r[0:4]
            [1, 3, 4, 7]
            sage: r = C(1-2*x^2)
            sage: r[0:4]
            [1, 0, -2, 0]
            sage: r[-1:4]             # not tested, python will not allow this!
            [0, 1, 0 -2, 0]
            sage: r = C((-2*x^3 + x^2 + 1)/(-2*x + 1))
            sage: r[0:5]              # handle ogf > 1
            [1, 2, 5, 8, 16]
            sage: r[-2]
            0
            sage: r = C((-2*x^3 + x^2 - x + 1)/(2*x^2 - 3*x + 1))
            sage: r[0:5]
            [1, 2, 5, 9, 17]
            sage: s=C((1-x)/(-x^2 - x + 1))
            sage: s[0:5]
            [1, 0, 1, 1, 2]
            sage: s=C((1+x^20+x^40)/(1-x^12)/(1-x^30))
            sage: s[0:20]
            [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0]
            sage: s=C(1/((1-x^2)*(1-x^6)*(1-x^8)*(1-x^12)))
            sage: s[999998]
            289362268629630
            sage: s = C.from_recurrence([1,1],[1,1,1])
            sage: s[0:5]
            [1, 1, 1, 2, 3]
            sage: s = C((1 - x)^-2); s[0:10]
            [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
            sage: s = C(x * (1 - x)^-2); s[0:10]
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
            sage: s = C(x^2 * (1 - x)^-2); s[0:10]
            [0, 0, 1, 2, 3, 4, 5, 6, 7, 8]
            sage: s = C(x^3 * (1 - x)^-2); s[0:10]
            [0, 0, 0, 1, 2, 3, 4, 5, 6, 7]
            sage: s = C(1/(1-x^1000)); s[10^18]
            1
            sage: s = C(1/(1-x^1000)); s[10^20]
            1

        REFERENCES:

        - [BM2021]_
        """
        if isinstance(key, slice):
            m = max(key.start, key.stop)
            return [self[ii] for ii in range(*key.indices(m + 1))]

        if isinstance(key, Integral):
            n = key - self._off
            if n < 0:
                return 0
            den = self.denominator()
            num = self.numerator()
            if self._off >= 0:
                num = num.shift(-self._off)
            else:
                den = den.shift(self._off)
            (quo, num) = num.quo_rem(den)
            if quo.degree() < n:
                wp = 0
            else:
                wp = quo[n]
            P = self.parent().polynomial_ring()
            x = P.gen()
            while n:
                nden = den(-x)
                num = P((num * nden).list()[n % 2::2])
                den = P((den * nden).list()[::2])
                n //= 2
            return wp + num[0] / den[0]

        raise TypeError("invalid argument type")

    def ogf(self):
        """
        Return the ordinary generating function associated with
        the CFiniteSequence.

        This is always a fraction of polynomials in the base ring.

        EXAMPLES::

            sage: C.<x> = CFiniteSequences(QQ)
            sage: r = C.from_recurrence([2],[1])
            sage: r.ogf()
            -1/2/(x - 1/2)
            sage: C(0).ogf()
            0
        """
        return self._ogf

    def numerator(self):
        r"""
        Return the numerator of the o.g.f of ``self``.

        EXAMPLES::

            sage: f = CFiniteSequence((2-x)/(1-x-x^2)); f
            C-finite sequence, generated by (x - 2)/(x^2 + x - 1)
            sage: f.numerator()
            x - 2
        """
        return self.ogf().numerator()

    def denominator(self):
        r"""
        Return the numerator of the o.g.f of ``self``.

        EXAMPLES::

            sage: f = CFiniteSequence((2-x)/(1-x-x^2)); f
            C-finite sequence, generated by (x - 2)/(x^2 + x - 1)
            sage: f.denominator()
            x^2 + x - 1
        """
        return self.ogf().denominator()

    def recurrence_repr(self) -> str:
        """
        Return a string with the recurrence representation of
        the C-finite sequence.

        OUTPUT: string

        EXAMPLES::

            sage: C.<x> = CFiniteSequences(QQ)
            sage: C((2-x)/(1-x-x^2)).recurrence_repr()
            'homogeneous linear recurrence with constant coefficients of degree 2: a(n+2) = a(n+1) + a(n), starting a(0...) = [2, 1]'
            sage: C(x/(1-x)^3).recurrence_repr()
            'homogeneous linear recurrence with constant coefficients of degree 3: a(n+3) = 3*a(n+2) - 3*a(n+1) + a(n), starting a(1...) = [1, 3, 6]'
            sage: C(1).recurrence_repr()
            'Finite sequence [1], offset 0'
            sage: r = C((-2*x^3 + x^2 - x + 1)/(2*x^2 - 3*x + 1))
            sage: r.recurrence_repr()
            'homogeneous linear recurrence with constant coefficients of degree 2: a(n+2) = 3*a(n+1) - 2*a(n), starting a(0...) = [1, 2, 5, 9]'
            sage: r = CFiniteSequence(x^3/(1-x-x^2))
            sage: r.recurrence_repr()
            'homogeneous linear recurrence with constant coefficients of degree 2: a(n+2) = a(n+1) + a(n), starting a(3...) = [1, 1, 2, 3]'
        """
        if self._deg == 0:
            return 'Finite sequence %s, offset %d' % (str(self._a), self._off)
        if self._c[0] == 1:
            cstr = 'a(n+%d) = a(n+%d)' % (self._deg, self._deg - 1)
        elif self._c[0] == -1:
            cstr = 'a(n+%d) = -a(n+%d)' % (self._deg, self._deg - 1)
        else:
            cstr = 'a(n+%d) = %s*a(n+%d)' % (self._deg, str(self._c[0]), self._deg - 1)
        for i in range(1, self._deg):
            j = self._deg - i - 1
            if self._c[i] < 0:
                if self._c[i] == -1:
                    cstr = cstr + ' - a(n+%d)' % (j,)
                else:
                    cstr = cstr + ' - %d*a(n+%d)' % (-(self._c[i]), j)
            elif self._c[i] > 0:
                if self._c[i] == 1:
                    cstr = cstr + ' + a(n+%d)' % (j,)
                else:
                    cstr = cstr + ' + %d*a(n+%d)' % (self._c[i], j)
        cstr = cstr.replace('+0', '')
        astr = ', starting a(%s...) = [' % str(self._off)
        maxwexp = self.numerator().quo_rem(self.denominator())[0].degree() + 1
        for i in range(maxwexp + self._deg):
            astr = astr + str(self[self._off + i]) + ', '
        astr = astr[:-2] + ']'
        return 'homogeneous linear recurrence with constant coefficients of degree ' + str(self._deg) + ': ' + cstr + astr

    def series(self, n):
        """
        Return the Laurent power series associated with the
        CFiniteSequence, with precision `n`.

        INPUT:

        - ``n`` -- nonnegative integer

        EXAMPLES::

            sage: C.<x> = CFiniteSequences(QQ)
            sage: r = C.from_recurrence([-1,2],[0,1])
            sage: s = r.series(4); s
            x + 2*x^2 + 3*x^3 + 4*x^4 + O(x^5)
            sage: type(s)
            <class 'sage.rings.laurent_series_ring_element.LaurentSeries'>
        """
        R = LaurentSeriesRing(QQ, 'x', default_prec=n)
        return R(self.ogf())

    def closed_form(self, n='n'):
        r"""
        Return a symbolic expression in ``n``, which equals the n-th term of
        the sequence.

        It is a well-known property of C-finite sequences ``a_n`` that they
        have a closed form of the type:

        .. MATH::

            a_n = \sum_{i=1}^d c_i(n) \cdot r_i^n,

        where ``r_i`` are the roots of the characteristic equation and
        ``c_i(n)`` is a polynomial (whose degree equals the multiplicity of
        ``r_i`` minus one).  This is a natural generalization of Binet's
        formula for Fibonacci numbers.  See, for instance, [KP2011, Theorem 4.1].

        Note that if the o.g.f. has a polynomial part, that is, if the
        numerator degree is not strictly less than the denominator degree,
        then this closed form holds only when ``n`` exceeds the degree of that
        polynomial part.  In that case, the returned expression will differ
        from the sequence for small ``n``.

        EXAMPLES::

            sage: CFiniteSequence(1/(1-x)).closed_form()
            1
            sage: CFiniteSequence(x^2/(1-x)).closed_form()
            1
            sage: CFiniteSequence(1/(1-x^2)).closed_form()
            1/2*(-1)^n + 1/2
            sage: CFiniteSequence(1/(1+x^3)).closed_form()
            1/3*(-1)^n + 1/3*(1/2*I*sqrt(3) + 1/2)^n + 1/3*(-1/2*I*sqrt(3) + 1/2)^n
            sage: CFiniteSequence(1/(1-x)/(1-2*x)/(1-3*x)).closed_form()
            9/2*3^n - 4*2^n + 1/2

        Binet's formula for the Fibonacci numbers::

            sage: CFiniteSequence(x/(1-x-x^2)).closed_form()
            sqrt(1/5)*(1/2*sqrt(5) + 1/2)^n - sqrt(1/5)*(-1/2*sqrt(5) + 1/2)^n
            sage: [_.subs(n=k).full_simplify() for k in range(6)]
            [0, 1, 1, 2, 3, 5]

            sage: CFiniteSequence((4*x+3)/(1-2*x-5*x^2)).closed_form()
            1/2*(sqrt(6) + 1)^n*(7*sqrt(1/6) + 3) - 1/2*(-sqrt(6) + 1)^n*(7*sqrt(1/6) - 3)

        Examples with multiple roots::

            sage: CFiniteSequence(x*(x^2+4*x+1)/(1-x)^5).closed_form()
            1/4*n^4 + 1/2*n^3 + 1/4*n^2
            sage: CFiniteSequence((1+2*x-x^2)/(1-x)^4/(1+x)^2).closed_form()
            1/12*n^3 - 1/8*(-1)^n*(n + 1) + 3/4*n^2 + 43/24*n + 9/8
            sage: CFiniteSequence(1/(1-x)^3/(1-2*x)^4).closed_form()
            4/3*(n^3 - 3*n^2 + 20*n - 36)*2^n + 1/2*n^2 + 19/2*n + 49
            sage: CFiniteSequence((x/(1-x-x^2))^2).closed_form()
            1/5*(n - sqrt(1/5))*(1/2*sqrt(5) + 1/2)^n + 1/5*(n + sqrt(1/5))*(-1/2*sqrt(5) + 1/2)^n
        """
        from sage.arith.misc import binomial
        from sage.rings.qqbar import QQbar

        from sage.symbolic.ring import SR
        n = SR(n)
        expr = SR.zero()

        R = FractionField(PolynomialRing(QQbar, self.parent().variable_name()))
        ogf = R(self.ogf())

        __, parts = ogf.partial_fraction_decomposition(decompose_powers=False)
        for part in parts:
            denom = part.denominator().factor()
            denom_base, denom_exp = denom[0]

            # denominator is of the form (x+b)^{m+1}
            m = denom_exp - 1
            b = denom_base.constant_coefficient()
            # check that the partial fraction decomposition was indeed done correctly
            # (that is, there is only one factor, of degree 1, and monic)
            assert len(denom) == 1 and len(denom_base.list()) == 2 and denom_base[1] == 1 and denom.unit() == 1

            r = SR((-1 / b).radical_expression())
            c = SR.zero()
            for k, a in enumerate(part.numerator()):
                a = -QQbar(a) if k % 2 else QQbar(a)
                bino = binomial(n + m - k, m)
                c += bino * SR((a * b**(k - m - 1)).radical_expression())

            expr += c.expand() * r**n

        return expr


class CFiniteSequences_generic(Parent, UniqueRepresentation):
    r"""
    The class representing the ring of C-Finite Sequences.

    TESTS::

        sage: C.<x> = CFiniteSequences(QQ)
        sage: from sage.rings.cfinite_sequence import CFiniteSequences_generic
        sage: isinstance(C,CFiniteSequences_generic)
        True
        sage: type(C)
        <class 'sage.rings.cfinite_sequence.CFiniteSequences_generic_with_category'>
        sage: C
        The ring of C-Finite sequences in x over Rational Field
    """

    Element = CFiniteSequence

    def __init__(self, polynomial_ring, category):
        r"""
        Create the ring of CFiniteSequences over ``base_ring``.

        INPUT:

        - ``base_ring`` -- the base ring for the o.g.f (either ``QQ`` or ``ZZ``)
        - ``names`` -- an iterable of variables (should contain only one variable)
        - ``category`` -- the category of the ring (default: ``Rings().Commutative()``)

        TESTS::

            sage: C.<y> = CFiniteSequences(QQ); C
            The ring of C-Finite sequences in y over Rational Field
            sage: C.<x> = CFiniteSequences(QQ); C
            The ring of C-Finite sequences in x over Rational Field
            sage: C.<x> = CFiniteSequences(ZZ); C
            The ring of C-Finite sequences in x over Integer Ring
            sage: C.<x,y> = CFiniteSequences(ZZ)
            Traceback (most recent call last):
            ...
            NotImplementedError: Multidimensional o.g.f. not implemented.
            sage: C.<x> = CFiniteSequences(CC)
            Traceback (most recent call last):
            ...
            ValueError: O.g.f. base not rational.
        """
        base_ring = polynomial_ring.base_ring()
        self._polynomial_ring = polynomial_ring
        self._fraction_field = FractionField(self._polynomial_ring)
        if category is None:
            category = Rings().Commutative()
        Parent.__init__(self, base_ring, names=self._polynomial_ring.gens(),
                        category=category)

    def _repr_(self):
        r"""
        Return the string representation of ``self``.

        EXAMPLES::

            sage: C.<x> = CFiniteSequences(QQ)
            sage: C
            The ring of C-Finite sequences in x over Rational Field
        """
        return "The ring of C-Finite sequences in {} over {}".format(self.gen(), self.base_ring())

    def _element_constructor_(self, ogf):
        r"""
        Construct a C-Finite Sequence.

        INPUT:

        - ``ogf`` -- the ordinary generating function, a fraction of
          polynomials over the rationals

        TESTS::

            sage: C.<x> = CFiniteSequences(QQ)
            sage: C((2-x)/(1-x-x^2))
            C-finite sequence, generated by (x - 2)/(x^2 + x - 1)
            sage: C(x/(1-x)^3)
            C-finite sequence, generated by -x/(x^3 - 3*x^2 + 3*x - 1)
            sage: C(x^2-4*x^5)
            Finite sequence [1, 0, 0, -4], offset = 2
            sage: C(x^2+3/x)
            Finite sequence [3, 0, 0, 1], offset = -1
            sage: C(1/x + 4/x^3)
            Finite sequence [4, 0, 1], offset = -3
            sage: P = LaurentPolynomialRing(QQ.fraction_field(), 'X')
            sage: X = P.gen()
            sage: C(1/(1-X))
            C-finite sequence, generated by -1/(x - 1)
            sage: C = CFiniteSequences(QQ)
            sage: C(x)
            Finite sequence [1], offset = 1
        """
        ogf = self.fraction_field()(ogf)
        return self.element_class(self, ogf)

    def ngens(self) -> int:
        r"""
        Return the number of generators of ``self``.

        EXAMPLES::

            sage: from sage.rings.cfinite_sequence import CFiniteSequences
            sage: C.<x> = CFiniteSequences(QQ)
            sage: C.ngens()
            1
        """
        return 1

    def gen(self, i=0):
        r"""
        Return the i-th generator of ``self``.

        INPUT:

        - ``i`` -- integer (default: 0)

        EXAMPLES::

            sage: C.<x> = CFiniteSequences(QQ)
            sage: C.gen()
            x
            sage: x == C.gen()
            True

        TESTS::

            sage: C.gen(2)
            Traceback (most recent call last):
            ...
            ValueError: The ring of C-Finite sequences in x over Rational Field has only one generator (i=0)
        """
        if i:
            raise ValueError("{} has only one generator (i=0)".format(self))
        return self.polynomial_ring().gen()

    def gens(self) -> tuple:
        """
        Return the generators of ``self``.

        EXAMPLES::

            sage: C.<x> = CFiniteSequences(QQ)
            sage: C.gens()
            (x,)
        """
        return (self.gen(0),)

    def an_element(self):
        r"""
        Return an element of C-Finite Sequences.

        OUTPUT: the Lucas sequence

        EXAMPLES::

            sage: C.<x> = CFiniteSequences(QQ)
            sage: C.an_element()
            C-finite sequence, generated by (x - 2)/(x^2 + x - 1)
        """
        x = self.gen()
        return self((2 - x) / (1 - x - x**2))

    def __contains__(self, x) -> bool:
        """
        Return ``True`` if x is an element of ``CFiniteSequences`` or
        canonically coerces to this ring.

        EXAMPLES::

            sage: C.<x> = CFiniteSequences(QQ)
            sage: x in C
            True
            sage: 1/x in C
            True
            sage: 5 in C
            True
            sage: pi in C
            False
            sage: Cy.<y> = CFiniteSequences(QQ)
            sage: y in C
            False
            sage: y in Cy
            True
        """
        return self.has_coerce_map_from(parent(x))

    def fraction_field(self):
        r"""
        Return the fraction field used to represent the elements of ``self``.

        EXAMPLES::

            sage: C.<x> = CFiniteSequences(QQ)
            sage: C.fraction_field()
            Fraction Field of Univariate Polynomial Ring in x over Rational Field
        """
        return self._fraction_field

    def polynomial_ring(self):
        r"""
        Return the polynomial ring used to represent the elements of ``self``.

        EXAMPLES::

            sage: C.<x> = CFiniteSequences(QQ)
            sage: C.polynomial_ring()
            Univariate Polynomial Ring in x over Rational Field
        """
        return self._polynomial_ring

    def _coerce_map_from_(self, S):
        """
        A coercion from `S` exists, if `S` coerces into ``self``'s fraction
        field.

        EXAMPLES::

            sage: C.<x> = CFiniteSequences(QQ)
            sage: C.has_coerce_map_from(C.fraction_field())
            True
            sage: C.has_coerce_map_from(QQ)
            True
            sage: C.has_coerce_map_from(QQ[x])
            True
            sage: C.has_coerce_map_from(ZZ)
            True
        """
        if self.fraction_field().has_coerce_map_from(S):
            return True

    def from_recurrence(self, coefficients, values):
        r"""
        Create a C-finite sequence given the coefficients `c` and
        starting values `a` of a homogeneous linear recurrence.

        .. MATH::

            a_{n+d} = c_0a_n + c_1a_{n+1} + \cdots + c_{d-1}a_{n+d-1}, \quad d\ge0.

        INPUT:

        - ``coefficients`` -- list of rationals
        - ``values`` -- start values, a list of rationals

        OUTPUT: a CFiniteSequence object

        EXAMPLES::

            sage: C.<x> = CFiniteSequences(QQ)
            sage: C.from_recurrence([1,1],[0,1])   # Fibonacci numbers
            C-finite sequence, generated by -x/(x^2 + x - 1)
            sage: C.from_recurrence([-1,2],[0,1])    # natural numbers
            C-finite sequence, generated by x/(x^2 - 2*x + 1)
            sage: r = C.from_recurrence([-1],[1])
            sage: s = C.from_recurrence([-1],[1,-1])
            sage: r == s
            True
            sage: r = C(x^3/(1-x-x^2))
            sage: s = C.from_recurrence([1,1],[0,0,0,1,1])
            sage: r == s
            True
            sage: C.from_recurrence(1,1)
            Traceback (most recent call last):
            ...
            ValueError: Wrong type for recurrence coefficient list.
        """
        if not isinstance(coefficients, list):
            raise ValueError("Wrong type for recurrence coefficient list.")
        if not isinstance(values, list):
            raise ValueError("Wrong type for recurrence start value list.")
        deg = len(coefficients)

        co = coefficients[::-1] + [0] * (len(values) - deg)
        R = self.polynomial_ring()
        den = R([-1] + co[:deg])
        num = R([-values[0]] +
                [-values[n] + sum(values[k] * co[n - 1 - k]
                                  for k in range(n))
                 for n in range(1, len(values))])
        return self(num / den)

    def guess(self, sequence, algorithm='sage'):
        """
        Return the minimal CFiniteSequence that generates the sequence.

        Assume the first value has index 0.

        INPUT:

        - ``sequence`` -- list of integers
        - ``algorithm`` -- string; one of
          - ``'sage'`` -- the default is to use Sage's matrix kernel function
          - ``'pari'`` -- use Pari's implementation of LLL
          - ``'bm'`` -- use Sage's Berlekamp-Massey algorithm

        OUTPUT: a CFiniteSequence, or 0 if none could be found

        With the default kernel method, trailing zeroes are chopped
        off before a guessing attempt. This may reduce the data
        below the accepted length of six values.

        EXAMPLES::

            sage: C.<x> = CFiniteSequences(QQ)
            sage: C.guess([1,2,4,8,16,32])
            C-finite sequence, generated by -1/2/(x - 1/2)
            sage: r = C.guess([1,2,3,4,5])
            Traceback (most recent call last):
            ...
            ValueError: sequence too short for guessing

        With Berlekamp-Massey, if an odd number of values is given, the last one is dropped.
        So with an odd number of values the result may not generate the last value::

            sage: r = C.guess([1,2,4,8,9], algorithm='bm'); r
            C-finite sequence, generated by -1/2/(x - 1/2)
            sage: r[0:5]
            [1, 2, 4, 8, 16]

        Using pari::

            sage: r = C.guess([1, 1, 1, 2, 2, 3, 4, 5, 7, 9, 12, 16, 21, 28], algorithm='pari'); r
            C-finite sequence, generated by (-x - 1)/(x^3 + x^2 - 1)
            sage: r[0:5]
            [1, 1, 1, 2, 2]
        """
        S = self.polynomial_ring()

        if algorithm == 'bm':
            from sage.matrix.berlekamp_massey import berlekamp_massey
            if len(sequence) < 2:
                raise ValueError('sequence too short for guessing')
            R = PowerSeriesRing(QQ, 'x')
            if len(sequence) % 2:
                sequence.pop()
            l = len(sequence) - 1
            denominator = S(berlekamp_massey(sequence).reverse())
            numerator = R(S(sequence) * denominator, prec=l).truncate()

            return CFiniteSequence(numerator / denominator)

        if algorithm == 'pari':
            if len(sequence) < 6:
                raise ValueError('sequence too short for guessing')
            pari("ggf(v)=local(l,m,p,q,B);l=length(v);B=l\\2;\
                if(B<3,return(0));m=matrix(B,B,x,y,v[x-y+B+1]);\
                q=qflll(m,4)[1];if(length(q)==0,return(0));\
                p=sum(k=1,B,x^(k-1)*q[k,1]);\
                q=Pol(Pol(vector(l,n,v[l-n+1]))*p+O(x^(B+1)));\
                if(polcoeff(p,0)<0,q=-q;p=-p);q=q/p;p=Ser(q+O(x^(l+1)));\
                for(m=1,l,if(polcoeff(p,m-1)!=v[m],return(0)));q")
            pari_guess = pari("ggf")(sequence)
            num = S(pari_guess.numerator().Vec().sage()[::-1])
            den = S(pari_guess.denominator().Vec().sage()[::-1])
            if num == 0:
                return 0
            return CFiniteSequence(num / den)

        from sage.matrix.constructor import matrix
        from sage.arith.misc import integer_ceil as ceil
        from numpy import trim_zeros
        seq = sequence[:]
        while seq and sequence[-1] == 0:
            seq.pop()
        l = len(seq)
        if l == 0:
            return 0
        if l < 6:
            raise ValueError('sequence too short for guessing')

        hl = ceil(ZZ(l) / 2)
        A = matrix([sequence[k: k + hl] for k in range(hl)])
        K = A.kernel()
        if K.dimension() == 0:
            return 0
        R = PolynomialRing(QQ, 'x')
        den = R(trim_zeros(K.basis()[-1].list()[::-1]))
        if den == 1:
            return 0
        offset = next((i for i, x in enumerate(sequence) if x), None)
        S = PowerSeriesRing(QQ, 'x', default_prec=l - offset)
        num = S(R(sequence) * den).truncate(ZZ(l) // 2 + 1)
        if num == 0 or sequence != S(num / den).list():
            return 0
        return CFiniteSequence(num / den)



# distutils: libraries = GSL_LIBRARIES
# distutils: library_dirs = GSL_LIBDIR
# distutils: include_dirs = GSL_INCDIR
from sage.libs.gsl.types cimport *

cdef extern from "gsl/gsl_eigen.h":

  ctypedef enum gsl_eigen_sort_t:
    GSL_EIGEN_SORT_VAL_ASC,
    GSL_EIGEN_SORT_VAL_DESC,
    GSL_EIGEN_SORT_ABS_ASC,
    GSL_EIGEN_SORT_ABS_DESC
  ctypedef struct gsl_eigen_symm_workspace
  ctypedef struct gsl_eigen_symmv_workspace
  ctypedef struct gsl_eigen_herm_workspace
  ctypedef struct gsl_eigen_hermv_workspace

  gsl_eigen_symm_workspace *  gsl_eigen_symm_alloc(size_t n)

  void  gsl_eigen_symm_free(gsl_eigen_symm_workspace * w)

  int  gsl_eigen_symm(gsl_matrix * A, gsl_vector * eval, gsl_eigen_symm_workspace * w)

  gsl_eigen_symmv_workspace *  gsl_eigen_symmv_alloc(size_t n)

  void  gsl_eigen_symmv_free(gsl_eigen_symmv_workspace * w)

  int  gsl_eigen_symmv(gsl_matrix * A, gsl_vector * eval, gsl_matrix * evec, gsl_eigen_symmv_workspace * w)

  gsl_eigen_herm_workspace *  gsl_eigen_herm_alloc(size_t n)

  void  gsl_eigen_herm_free(gsl_eigen_herm_workspace * w)

  int  gsl_eigen_herm(gsl_matrix_complex * A, gsl_vector * eval, gsl_eigen_herm_workspace * w)

  gsl_eigen_hermv_workspace *  gsl_eigen_hermv_alloc(size_t n)

  void  gsl_eigen_hermv_free(gsl_eigen_hermv_workspace * w)

  int  gsl_eigen_hermv(gsl_matrix_complex * A, gsl_vector * eval, gsl_matrix_complex * evec, gsl_eigen_hermv_workspace * w)

  int  gsl_eigen_symmv_sort(gsl_vector * eval, gsl_matrix * evec, gsl_eigen_sort_t sort_type)

  int  gsl_eigen_hermv_sort(gsl_vector * eval, gsl_matrix_complex * evec, gsl_eigen_sort_t sort_type)




# sage.doctest: needs sage.libs.singular
"""
Base class of curves

This module defines the base class of curves in Sage.

Curves in Sage are reduced subschemes of dimension 1 of an ambient space. The ambient space
is either an affine space or a projective space.

EXAMPLES::

    sage: A.<x,y,z> = AffineSpace(QQ, 3)
    sage: C = Curve([x - y, z - 2])
    sage: C
    Affine Curve over Rational Field defined by x - y, z - 2
    sage: C.dimension()
    1

AUTHORS:

- William Stein (2005)
"""
# ****************************************************************************
#       Copyright (C) 2005 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.misc.latex import latex

from sage.categories.finite_fields import FiniteFields
from sage.categories.fields import Fields

from sage.schemes.generic.algebraic_scheme import AlgebraicScheme_subscheme
from sage.schemes.generic.divisor_group import DivisorGroup
from sage.schemes.generic.divisor import Divisor_curve

from sage.rings.integer import Integer


class Curve_generic(AlgebraicScheme_subscheme):
    r"""
    Generic curve class.

    EXAMPLES::

        sage: A.<x,y,z> = AffineSpace(QQ, 3)
        sage: C = Curve([x - y, z - 2])
        sage: loads(C.dumps()) == C
        True
    """
    def _repr_(self):
        """
        Return a string representation of this curve.

        EXAMPLES::

            sage: A.<x,y,z> = AffineSpace(QQ, 3)
            sage: C = Curve([x - y, z - 2])
            sage: C
            Affine Curve over Rational Field defined by x - y, z - 2

            sage: P.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: C = Curve(x - y)
            sage: C
            Projective Plane Curve over Rational Field defined by x - y
        """
        if self.defining_ideal().is_zero() and self.ambient_space().dimension() == 1:
            return "{} Line over {}".format(self._repr_type(), self.base_ring())
        polys = ', '.join(str(x) for x in self.defining_polynomials())
        return "{} Curve over {} defined by {}".format(self._repr_type(),
                                                       self.base_ring(), polys)

    def _repr_type(self) -> str:
        r"""
        Return a string representation of the type of this curve.

        EXAMPLES::

            sage: P.<x,y,z,w> = ProjectiveSpace(QQ, 3)
            sage: C = Curve([w^3 - z^3, w*x - x^2])
            sage: from sage.schemes.curves.curve import Curve_generic
            sage: Curve_generic._repr_type(C)
            'Generic'
        """
        return "Generic"

    def _latex_(self):
        r"""
        Return a latex representation of this curve.

        EXAMPLES::

            sage: x,y,z = PolynomialRing(QQ, 3, names='x,y,z').gens()
            sage: C = Curve(y^2*z - x^3 - 17*x*z^2 + y*z^2)
            sage: latex(C)
            \text{Projective Plane curve over $\Bold{Q}$
            defined by $-x^{3} + y^{2} z - 17 x z^{2} + y z^{2}$}

            sage: A2 = AffineSpace(2, QQ, names=['x','y'])
            sage: x, y = A2.coordinate_ring().gens()
            sage: C = Curve(y^2 - x^3 - 17*x + y)
            sage: latex(C)
            \text{Affine Plane curve over $\Bold{Q}$
            defined by $-x^{3} + y^{2} - 17 x + y$}
        """
        if (self.defining_ideal().is_zero()
                and self.ambient_space().dimension() == 1):
            ambient_type, ring = self._repr_type(), latex(self.base_ring())
            return fr"\text{{{ambient_type} line over ${ring}$}}"
        ambient_type, ring = self._repr_type(), latex(self.base_ring())
        polys = ', '.join(f'${latex(p)}$' for p in self.defining_polynomials())
        return fr"\text{{{ambient_type} curve over ${ring}$ defined by {polys}}}"

    def dimension(self):
        r"""
        Return the dimension of the curve.

        Curves have dimension one by definition.

        EXAMPLES::

            sage: x = polygen(QQ)
            sage: C = HyperellipticCurve(x^7 + x^4 + x)
            sage: C.dimension()
            1
            sage: from sage.schemes.projective.projective_subscheme import AlgebraicScheme_subscheme_projective
            sage: AlgebraicScheme_subscheme_projective.dimension(C)
            1
        """
        return Integer(1)

    def defining_polynomial(self):
        """
        Return the defining polynomial of the curve.

        EXAMPLES::

            sage: x,y,z = PolynomialRing(QQ, 3, names='x,y,z').gens()
            sage: C = Curve(y^2*z - x^3 - 17*x*z^2 + y*z^2)
            sage: C.defining_polynomial()
            -x^3 + y^2*z - 17*x*z^2 + y*z^2
        """
        return self.defining_polynomials()[0]

    def divisor_group(self, base_ring=None):
        r"""
        Return the divisor group of the curve.

        INPUT:

        - ``base_ring`` -- the base ring of the divisor group; usually, this is
          `\ZZ` (default) or `\QQ`

        OUTPUT: the divisor group of the curve

        EXAMPLES::

            sage: x,y,z = PolynomialRing(QQ, 3, names='x,y,z').gens()
            sage: C  = Curve(y^2*z - x^3 - 17*x*z^2 + y*z^2)
            sage: Cp = Curve(y^2*z - x^3 - 17*x*z^2 + y*z^2)
            sage: C.divisor_group() is Cp.divisor_group()
            True
        """
        return DivisorGroup(self, base_ring)

    def divisor(self, v, base_ring=None, check=True, reduce=True):
        r"""
        Return the divisor specified by ``v``.

        .. WARNING::

            The coefficients of the divisor must be in the base ring
            and the terms must be reduced. If you set ``check=False``
            and/or ``reduce=False`` it is your responsibility to pass
            a valid object ``v``.

        EXAMPLES::

            sage: x,y,z = PolynomialRing(QQ, 3, names='x,y,z').gens()
            sage: C = Curve(y^2*z - x^3 - 17*x*z^2 + y*z^2)
            sage: p1 = C(0, -1, 1)
            sage: p2 = C(0, 0, 1)
            sage: p3 = C(0, 1, 0)
            sage: C.divisor([(1, p1), (-1, p2), (2, p3)])
            (x, y + z) - (x, y) + 2*(x, z)
        """
        return Divisor_curve(v, check=check, reduce=reduce, parent=self.divisor_group(base_ring))

    def genus(self):
        """
        Return the geometric genus of the curve.

        EXAMPLES::

            sage: x,y,z = PolynomialRing(QQ, 3, names='x,y,z').gens()
            sage: C = Curve(y^2*z - x^3 - 17*x*z^2 + y*z^2)
            sage: C.genus()
            1
        """
        return self.geometric_genus()

    def geometric_genus(self):
        r"""
        Return the geometric genus of the curve.

        EXAMPLES:

        Examples of projective curves::

            sage: P2 = ProjectiveSpace(2, GF(5), names=['x','y','z'])
            sage: x, y, z = P2.coordinate_ring().gens()
            sage: C = Curve(y^2*z - x^3 - 17*x*z^2 + y*z^2)
            sage: C.geometric_genus()
            1
            sage: C = Curve(y^2*z - x^3)
            sage: C.geometric_genus()
            0
            sage: C = Curve(x^10 + y^7*z^3 + z^10)
            sage: C.geometric_genus()
            3

        Examples of affine curves::

            sage: x, y = PolynomialRing(GF(5), 2, 'xy').gens()
            sage: C = Curve(y^2 - x^3 - 17*x + y)
            sage: C.geometric_genus()
            1
            sage: C = Curve(y^2 - x^3)
            sage: C.geometric_genus()
            0
            sage: C = Curve(x^10 + y^7 + 1)
            sage: C.geometric_genus()
            3

        .. WARNING::

            Geometric genus is only defined for `geometrically irreducible curve
            <https://stacks.math.columbia.edu/tag/0BYE>`_. This method does not
            check the condition. You may get a nonsensical result if the curve is
            not geometrically irreducible::

                sage: P2.<x,y,z> = ProjectiveSpace(QQ, 2)
                sage: C = Curve(x^2 + y^2, P2)
                sage: C.geometric_genus()  # nonsense!
                -1
        """
        try:
            return self._genus
        except AttributeError:
            raise NotImplementedError

    def union(self, other):
        """
        Return the union of ``self`` and ``other``.

        EXAMPLES::

            sage: x,y,z = PolynomialRing(QQ, 3, names='x,y,z').gens()
            sage: C1 = Curve(z - x)
            sage: C2 = Curve(y - x)
            sage: C1.union(C2).defining_polynomial()
            x^2 - x*y - x*z + y*z
        """
        from .constructor import Curve
        return Curve(AlgebraicScheme_subscheme.union(self, other))

    __add__ = union

    def singular_subscheme(self):
        r"""
        Return the subscheme of singular points of this curve.

        OUTPUT: a subscheme in the ambient space of this curve

        EXAMPLES::

            sage: A.<x,y> = AffineSpace(CC, 2)
            sage: C = Curve([y^4 - 2*x^5 - x^2*y], A)
            sage: C.singular_subscheme()
            Closed subscheme of Affine Space of dimension 2 over Complex Field
             with 53 bits of precision defined by:
              (-2.00000000000000)*x^5 + y^4 - x^2*y,
              (-10.0000000000000)*x^4 + (-2.00000000000000)*x*y,
              4.00000000000000*y^3 - x^2

        ::

            sage: P.<x,y,z,w> = ProjectiveSpace(QQ, 3)
            sage: C = Curve([y^8 - x^2*z*w^5, w^2 - 2*y^2 - x*z], P)
            sage: C.singular_subscheme()
            Closed subscheme of Projective Space of dimension 3
             over Rational Field defined by:
              y^8 - x^2*z*w^5,
              -2*y^2 - x*z + w^2,
              -x^3*y*z^4 + 3*x^2*y*z^3*w^2 - 3*x*y*z^2*w^4 + 8*x*y*z*w^5 + y*z*w^6,
              x^2*z*w^5,
              -5*x^2*z^2*w^4 - 4*x*z*w^6,
              x^4*y*z^3 - 3*x^3*y*z^2*w^2 + 3*x^2*y*z*w^4 - 4*x^2*y*w^5 - x*y*w^6,
              -2*x^3*y*z^3*w + 6*x^2*y*z^2*w^3 - 20*x^2*y*z*w^4
               - 6*x*y*z*w^5 + 2*y*w^7,
              -5*x^3*z*w^4 - 2*x^2*w^6
        """
        return self.ambient_space().subscheme(self.Jacobian())

    def singular_points(self, F=None):
        r"""
        Return the set of singular points of this curve.

        INPUT:

        - ``F`` -- (default: ``None``) field over which to find the singular
          points; if not given, the base ring of this curve is used

        OUTPUT: list of points in the ambient space of this curve

        EXAMPLES::

            sage: A.<x,y,z> = AffineSpace(QQ, 3)
            sage: C = Curve([y^2 - x^5, x - z], A)
            sage: C.singular_points()
            [(0, 0, 0)]

        ::

            sage: # needs sage.rings.number_field
            sage: R.<a> = QQ[]
            sage: K.<b> = NumberField(a^8 - a^4 + 1)
            sage: P.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: C = Curve([359/12*x*y^2*z^2 + 2*y*z^4 + 187/12*y^3*z^2 + x*z^4
            ....:            + 67/3*x^2*y*z^2 + 117/4*y^5 + 9*x^5 + 6*x^3*z^2
            ....:            + 393/4*x*y^4 + 145*x^2*y^3 + 115*x^3*y^2 + 49*x^4*y], P)
            sage: sorted(C.singular_points(K), key=str)
            [(-1/2*b^5 - 1/2*b^3 + 1/2*b - 1 : 1 : 0),
             (-2/3*b^4 + 1/3 : 0 : 1),
             (-b^6 : b^6 : 1),
             (1/2*b^5 + 1/2*b^3 - 1/2*b - 1 : 1 : 0),
             (2/3*b^4 - 1/3 : 0 : 1),
             (b^6 : -b^6 : 1)]
        """
        if F is None:
            if self.base_ring() not in Fields():
                raise TypeError("curve must be defined over a field")
            F = self.base_ring()
        elif F not in Fields():
            raise TypeError("(=%s) must be a field" % F)
        X = self.singular_subscheme()
        return [self.point(p, check=False) for p in X.rational_points(F=F)]

    def is_singular(self, P=None) -> bool:
        r"""
        Return whether ``P`` is a singular point of this curve, or if no point
        is passed, whether this curve is singular or not.

        This just uses the is_smooth function for algebraic subschemes.

        INPUT:

        - ``P`` -- (default: ``None``) a point on this curve

        OUTPUT:

        boolean; if a point ``P`` is provided, and if ``P`` lies on this
        curve, returns ``True`` if ``P`` is a singular point of this curve, and
        ``False`` otherwise. If no point is provided, returns ``True`` or False
        depending on whether this curve is or is not singular, respectively.

        EXAMPLES::

            sage: P.<x,y,z,w> = ProjectiveSpace(QQ, 3)
            sage: C = P.curve([y^2 - x^2 - z^2, z - w])
            sage: C.is_singular()
            False

        ::

            sage: A.<x,y,z> = AffineSpace(GF(11), 3)
            sage: C = A.curve([y^3 - z^5, x^5 - y + 1])
            sage: Q = A([7,0,0])
            sage: C.is_singular(Q)
            True
        """
        return not self.is_smooth(P)

    def intersects_at(self, C, P):
        r"""
        Return whether the point ``P`` is or is not in the intersection of this
        curve with the curve ``C``.

        INPUT:

        - ``C`` -- a curve in the same ambient space as this curve

        - ``P`` -- a point in the ambient space of this curve

        EXAMPLES::

            sage: P.<x,y,z,w> = ProjectiveSpace(QQ, 3)
            sage: C = Curve([x^2 - z^2, y^3 - w*x^2], P)
            sage: D = Curve([w^2 - 2*x*y + z^2, y^2 - w^2], P)
            sage: Q1 = P([1,1,-1,1])
            sage: C.intersects_at(D, Q1)
            True
            sage: Q2 = P([0,0,1,-1])
            sage: C.intersects_at(D, Q2)
            False

        ::

            sage: A.<x,y> = AffineSpace(GF(13), 2)
            sage: C = Curve([y + 12*x^5 + 3*x^3 + 7], A)
            sage: D = Curve([y^2 + 7*x^2 + 8], A)
            sage: Q1 = A([9,6])
            sage: C.intersects_at(D, Q1)
            True
            sage: Q2 = A([3,7])
            sage: C.intersects_at(D, Q2)
            False
        """
        if C.ambient_space() != self.ambient_space():
            raise TypeError("(=%s) must be a curve in the same ambient space as (=%s)" % (C, self))
        if not isinstance(C, Curve_generic):
            raise TypeError("(=%s) must be a curve" % C)
        try:
            P = self.ambient_space()(P)
        except TypeError:
            raise TypeError("(=%s) must be a point in the ambient space of this curve" % P)
        try:
            P = self(P)
            P = C(P)
        except TypeError:
            return False
        return True

    def intersection_points(self, C, F=None):
        r"""
        Return the points in the intersection of this curve and the curve ``C``.

        If the intersection of these two curves has dimension greater than
        zero, and if the base ring of this curve is not a finite field, then an
        error is returned.

        INPUT:

        - ``C`` -- a curve in the same ambient space as this curve

        - ``F`` -- (default: ``None``) field over which to compute the
          intersection points; if not specified, the base ring of this curve is
          used

        OUTPUT: list of points in the ambient space of this curve

        EXAMPLES::

            sage: # needs sage.rings.number_field
            sage: R.<a> = QQ[]
            sage: K.<b> = NumberField(a^2 + a + 1)
            sage: P.<x,y,z,w> = ProjectiveSpace(QQ, 3)
            sage: C = Curve([y^2 - w*z, w^3 - y^3], P)
            sage: D = Curve([x*y - w*z, z^3 - y^3], P)
            sage: C.intersection_points(D, F=K)
            [(-b - 1 : -b - 1 : b : 1), (b : b : -b - 1 : 1),
             (1 : 0 : 0 : 0), (1 : 1 : 1 : 1)]

        ::

            sage: A.<x,y> = AffineSpace(GF(7), 2)
            sage: C = Curve([y^3 - x^3], A)
            sage: D = Curve([-x*y^3 + y^4 - 2*x^3 + 2*x^2*y], A)
            sage: C.intersection_points(D)
            [(0, 0), (1, 1), (2, 2), (3, 3), (4, 4), (5, 3), (5, 5), (5, 6), (6, 6)]

        ::

            sage: A.<x,y> = AffineSpace(QQ, 2)
            sage: C = Curve([y^3 - x^3], A)
            sage: D = Curve([-x*y^3 + y^4 - 2*x^3 + 2*x^2*y], A)
            sage: C.intersection_points(D)
            Traceback (most recent call last):
            ...
            NotImplementedError: the intersection must have dimension zero or
            (=Rational Field) must be a finite field
        """
        if C.ambient_space() != self.ambient_space():
            raise TypeError("(=%s) must be a curve in the same ambient space as (=%s)" % (C, self))
        if not isinstance(C, Curve_generic):
            raise TypeError("(=%s) must be a curve" % C)
        X = self.intersection(C)
        if F is None:
            F = self.base_ring()
        if X.dimension() == 0 or F in FiniteFields():
            return X.rational_points(F=F)
        raise NotImplementedError("the intersection must have dimension "
                                  "zero or (={}) must be a finite field".format(F))

    def change_ring(self, R):
        r"""
        Return a new curve which is this curve coerced to ``R``.

        INPUT:

        - ``R`` -- ring or embedding

        OUTPUT: a new curve which is this curve coerced to ``R``

        EXAMPLES::

            sage: P.<x,y,z,w> = ProjectiveSpace(QQ, 3)
            sage: C = Curve([x^2 - y^2, z*y - 4/5*w^2], P)
            sage: C.change_ring(QuadraticField(-1))                                     # needs sage.rings.number_field
            Projective Curve over Number Field in a with defining polynomial x^2 + 1
             with a = 1*I defined by x^2 - y^2, y*z - 4/5*w^2

        ::

            sage: # needs sage.rings.number_field
            sage: R.<a> = QQ[]
            sage: K.<b> = NumberField(a^3 + a^2 - 1)
            sage: A.<x,y> = AffineSpace(K, 2)
            sage: C = Curve([K.0*x^2 - x + y^3 - 11], A)
            sage: L = K.embeddings(QQbar)
            sage: set_verbose(-1)  # suppress warnings for slow computation
            sage: C.change_ring(L[0])
            Affine Plane Curve over Algebraic Field defined
             by y^3 + (-0.8774388331233464? - 0.744861766619745?*I)*x^2 - x - 11

        ::

            sage: P.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: C = P.curve([y*x - 18*x^2 + 17*z^2])
            sage: C.change_ring(GF(17))
            Projective Plane Curve over Finite Field of size 17 defined by -x^2 + x*y
        """
        new_AS = self.ambient_space().change_ring(R)
        I = [f.change_ring(R) for f in self.defining_polynomials()]
        return new_AS.curve(I)



"""
Ideals of Finite Dimensional Algebras

It is necessary to use algebras in the category of associative algebras.
"""

# ****************************************************************************
#  Copyright (C) 2011 Johan Bosman <johan.g.bosman@gmail.com>
#  Copyright (C) 2011, 2013 Peter Bruin <peter.bruin@math.uzh.ch>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from functools import reduce

from sage.algebras.finite_dimensional_algebras.finite_dimensional_algebra_element import (
    FiniteDimensionalAlgebraElement,
)
from sage.matrix.constructor import matrix
from sage.misc.cachefunc import cached_method
from sage.rings.ideal import Ideal_generic
from sage.structure.element import Matrix, parent
from sage.structure.richcmp import op_EQ, op_GE, op_GT, op_LE, op_LT, op_NE


class FiniteDimensionalAlgebraIdeal(Ideal_generic):
    """
    An ideal of a :class:`FiniteDimensionalAlgebra`.

    INPUT:

    - ``A`` -- a finite-dimensional algebra
    - ``gens`` -- the generators of this ideal
    - ``given_by_matrix`` -- boolean (default: ``False``); whether the basis
      matrix is given by ``gens``

    EXAMPLES::

        sage: cat = CommutativeAlgebras(GF(3)).FiniteDimensional().WithBasis()
        sage: A = FiniteDimensionalAlgebra(GF(3), [Matrix([[1, 0], [0, 1]]),
        ....:                                      Matrix([[0, 1], [0, 0]])],
        ....:                              category=cat)
        sage: A.ideal(A([0,1]))
        Ideal (e1) of Finite-dimensional algebra of degree 2 over Finite Field of size 3
    """
    def __init__(self, A, gens=None, given_by_matrix=False) -> bool:
        """
        EXAMPLES::

            sage: cat = CommutativeAlgebras(GF(3)).FiniteDimensional().WithBasis()
            sage: A = FiniteDimensionalAlgebra(GF(3), [Matrix([[1, 0], [0, 1]]),
            ....:                                      Matrix([[0, 1], [0, 0]])],
            ....:                              category=cat)
            sage: I = A.ideal(A([0,1]))
            sage: TestSuite(I).run(skip='_test_category')  # Currently ideals are not using the category framework
        """
        k = A.base_ring()
        n = A.degree()
        if given_by_matrix:
            self._basis_matrix = gens
            gens = gens.rows()
        elif gens is None:
            self._basis_matrix = matrix(k, 0, n)
        elif isinstance(gens, (list, tuple)):
            B = [FiniteDimensionalAlgebraIdeal(A, x).basis_matrix() for x in gens]
            B = reduce(lambda x, y: x.stack(y), B, matrix(k, 0, n))
            self._basis_matrix = B.echelon_form().image().basis_matrix()
        elif isinstance(gens, Matrix):
            gens = FiniteDimensionalAlgebraElement(A, gens)
        elif isinstance(gens, FiniteDimensionalAlgebraElement):
            gens = gens.vector()
            B = matrix([(gens * b).list() for b in A.table()])
            self._basis_matrix = B.echelon_form().image().basis_matrix()
        Ideal_generic.__init__(self, A, gens)

    def _richcmp_(self, other, op):
        r"""
        Comparisons.

        TESTS::

            sage: cat = CommutativeAlgebras(GF(3)).FiniteDimensional().WithBasis()
            sage: A = FiniteDimensionalAlgebra(GF(3), [Matrix([[1, 0], [0, 1]]),
            ....:                                      Matrix([[0, 1], [0, 0]])],
            ....:                              category=cat)
            sage: I = A.ideal(A([1,1]))
            sage: J = A.ideal(A([0,1]))
            sage: I == J
            False
            sage: I == I
            True
            sage: I == I + J
            True

            sage: A2 = FiniteDimensionalAlgebra(GF(3), [Matrix([[1, 0], [0, 1]]),
            ....:                                       Matrix([[0, 1], [0, 0]])],
            ....:     category=cat)
            sage: A is A2
            True
            sage: A == A2
            True
            sage: I2 = A.ideal(A([1,1]))
            sage: I == I2
            True

            sage: I != J, I != I, I != I+J
            (True, False, False)
            sage: I <= J, I <= I, I <= I+J
            (False, True, True)
            sage: I < J, I < I, I < I+J
            (False, False, False)
            sage: I >= J, I >= I, I >= I+J
            (True, True, True)
            sage: I > J, I > I, I > I+J
            (True, False, False)

            sage: I = A.ideal(A([1,1]))
            sage: J = A.ideal(A([0,1]))
            sage: I != J
            True
            sage: I != I
            False
            sage: I != I + J
            False
        """
        if self.basis_matrix() == other.basis_matrix():
            return op == op_EQ or op == op_LE or op == op_GE
        if op == op_EQ:
            return False
        if op == op_NE:
            return True
        if op == op_LE or op == op_LT:
            return self.vector_space().is_subspace(other.vector_space())
        if op == op_GE or op == op_GT:
            return other.vector_space().is_subspace(self.vector_space())

    def __contains__(self, elt) -> bool:
        """
        EXAMPLES::

            sage: cat = CommutativeAlgebras(GF(3)).FiniteDimensional().WithBasis()
            sage: A = FiniteDimensionalAlgebra(GF(3), [Matrix([[1, 0], [0, 1]]),
            ....:                                      Matrix([[0, 1], [0, 0]])],
            ....:                              category=cat)
            sage: J = A.ideal(A([0,1]))
            sage: A([0,1]) in J
            True
            sage: A([1,0]) in J
            False
        """
        if self.ring() is not parent(elt):
            return False
        return elt.vector() in self.vector_space()

    def basis_matrix(self):
        """
        Return the echelonized matrix whose rows form a basis of ``self``.

        EXAMPLES::

            sage: cat = CommutativeAlgebras(GF(3)).FiniteDimensional().WithBasis()
            sage: A = FiniteDimensionalAlgebra(GF(3), [Matrix([[1, 0], [0, 1]]),
            ....:                                      Matrix([[0, 1], [0, 0]])],
            ....:                              category=cat)
            sage: I = A.ideal(A([1,1]))
            sage: I.basis_matrix()
            [1 0]
            [0 1]
        """
        return self._basis_matrix

    @cached_method
    def vector_space(self):
        """
        Return ``self`` as a vector space.

        EXAMPLES::

            sage: cat = CommutativeAlgebras(GF(3)).FiniteDimensional().WithBasis()
            sage: A = FiniteDimensionalAlgebra(GF(3), [Matrix([[1, 0], [0, 1]]),
            ....:                                      Matrix([[0, 1], [0, 0]])],
            ....:                              category=cat)
            sage: I = A.ideal(A([1,1]))
            sage: I.vector_space()
            Vector space of degree 2 and dimension 2 over Finite Field of size 3
            Basis matrix:
            [1 0]
            [0 1]
        """
        return self.basis_matrix().image()



r"""
Introduction

Sage has a wide support for 3D graphics, from basic shapes to implicit and
parametric plots.

The following graphics functions are supported:


-  :func:`~plot3d` -- plot a 3d function

-  :func:`~sage.plot.plot3d.parametric_plot3d.parametric_plot3d` -- a parametric three-dimensional space curve or surface

-  :func:`~sage.plot.plot3d.revolution_plot3d.revolution_plot3d` -- a plot of a revolved curve

-  :func:`~sage.plot.plot3d.plot_field3d.plot_vector_field3d` -- a plot of a 3d vector field

-  :func:`~sage.plot.plot3d.implicit_plot3d.implicit_plot3d` -- a plot of an isosurface of a function

-  :func:`~sage.plot.plot3d.list_plot3d.list_plot3d`- a 3-dimensional plot of a surface defined by a list of points in 3-dimensional space

-  :func:`~sage.plot.plot3d.list_plot3d.list_plot3d_matrix` -- a 3-dimensional plot of a surface defined by a matrix defining points in 3-dimensional space

-  :func:`~sage.plot.plot3d.list_plot3d.list_plot3d_array_of_arrays`- A 3-dimensional plot of a surface defined by a list of lists defining points in 3-dimensional space

-  :func:`~sage.plot.plot3d.list_plot3d.list_plot3d_tuples` -- a 3-dimensional plot of a surface defined by a list of points in 3-dimensional space

The following classes for basic shapes are supported:


-  :class:`~sage.plot.plot3d.shapes.Box` -- a box given its three magnitudes

-  :class:`~sage.plot.plot3d.shapes.Cone` -- a cone, with base in the xy-plane pointing up the z-axis

-  :class:`~sage.plot.plot3d.shapes.Cylinder` -- a cylinder, with base in the xy-plane pointing up the z-axis

-  :class:`~sage.plot.plot3d.shapes2.Line` -- a 3d line joining a sequence of points

-  :class:`~sage.plot.plot3d.shapes.Sphere` -- a sphere centered at the origin

-  :class:`~sage.plot.plot3d.shapes.Text` -- a text label attached to a point in 3d space

-  :class:`~sage.plot.plot3d.shapes.Torus` -- a 3d torus

-  :class:`~sage.plot.plot3d.shapes2.Point` -- a position in 3d, represented by a sphere of fixed size


The following plotting functions for basic shapes are supported


-  :func:`~sage.plot.plot3d.shapes.ColorCube` -- a cube with given size and sides with given colors

-  :func:`~sage.plot.plot3d.shapes.LineSegment` -- a line segment, which is drawn as a cylinder from start to end with given radius

-  :func:`~sage.plot.plot3d.shapes2.line3d` -- a 3d line joining a sequence of points

-  :func:`~sage.plot.plot3d.shapes.arrow3d` -- a 3d arrow

-  :func:`~sage.plot.plot3d.shapes2.point3d` -- a point or list of points in 3d space

-  :func:`~sage.plot.plot3d.shapes2.bezier3d` -- a 3d bezier path

-  :func:`~sage.plot.plot3d.shapes2.frame3d` -- a frame in 3d

-  :func:`~sage.plot.plot3d.shapes2.frame_labels` -- labels for a given frame in 3d

-  :func:`~sage.plot.plot3d.shapes2.polygon3d` -- draw a polygon in 3d

-  :func:`~sage.plot.plot3d.shapes2.polygons3d` -- draw the union of several polygons in 3d

-  :func:`~sage.plot.plot3d.shapes2.ruler` -- draw a ruler in 3d, with major and minor ticks

-  :func:`~sage.plot.plot3d.shapes2.ruler_frame` -- draw a frame made of 3d rulers, with major and minor ticks

-  :func:`~sage.plot.plot3d.shapes2.sphere` -- plot of a sphere given center and radius

-  :func:`~sage.plot.plot3d.shapes2.text3d` -- 3d text

Sage also supports platonic solids with the following functions:


-  :func:`~sage.plot.plot3d.platonic.tetrahedron`

-  :func:`~sage.plot.plot3d.platonic.cube`

-  :func:`~sage.plot.plot3d.platonic.octahedron`

-  :func:`~sage.plot.plot3d.platonic.dodecahedron`

-  :func:`~sage.plot.plot3d.platonic.icosahedron`

Different viewers are supported: a web-based interactive viewer using the
Three.js JavaScript library (the default), Jmol, and the Tachyon ray tracer.
The viewer is invoked by adding the keyword argument
``viewer='threejs'`` (respectively ``'jmol'`` or ``'tachyon'``)
to the command ``show()`` on any three-dimensional graphic.


-  :class:`~sage.plot.plot3d.tachyon.Tachyon` -- create a scene the can be rendered using the Tachyon ray tracer

-  :class:`~sage.plot.plot3d.tachyon.Axis_aligned_box` -- box with axis-aligned edges with the given min and max coordinates

-  :class:`~sage.plot.plot3d.tachyon.Cylinder` -- an infinite cylinder

-  :class:`~sage.plot.plot3d.tachyon.FCylinder` -- a finite cylinder

-  :class:`~sage.plot.plot3d.tachyon.FractalLandscape`- axis-aligned fractal landscape

-  :class:`~sage.plot.plot3d.tachyon.Light` -- represents lighting objects

-  :class:`~sage.plot.plot3d.tachyon.ParametricPlot` -- parametric plot routines

-  :class:`~sage.plot.plot3d.tachyon.Plane` -- an infinite plane

-  :class:`~sage.plot.plot3d.tachyon.Ring` -- an annulus of zero thickness

-  :class:`~sage.plot.plot3d.tachyon.Sphere`- a sphere

-  :class:`~sage.plot.plot3d.tachyon.TachyonSmoothTriangle` -- a triangle along with a normal vector, which is used for smoothing

-  :class:`~sage.plot.plot3d.tachyon.TachyonTriangle` -- basic triangle class

-  :class:`~sage.plot.plot3d.tachyon.TachyonTriangleFactory` -- class to produce triangles of various rendering types

-  :class:`~sage.plot.plot3d.tachyon.Texfunc` -- creates a texture function

-  :class:`~sage.plot.plot3d.tachyon.Texture` -- stores texture information

-  :func:`~sage.plot.plot3d.tachyon.tostr` -- converts vector information to a space-separated string
"""



"""
Lie Algebras
"""

# *****************************************************************************
#       Copyright (C) 2013 Travis Scrimshaw <tscrim at ucdavis.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# *****************************************************************************
from sage.misc.lazy_import import lazy_import

lazy_import('sage.algebras.lie_algebras.lie_algebra', 'LieAlgebra')
# from kac_moody import KacMoodyAlgebra
lazy_import('sage.algebras.lie_algebras', 'examples', 'lie_algebras')
del lazy_import



from collections.abc import Callable
from typing import Any

def dict_key(o: Any) -> Any:
    ...

def cache_key(o: Any) -> Any:
    ...

def cached_method(f=None, name: str | None = None, key=None, do_pickle: bool = False) -> CachedMethod:
    ...

def cached_function(f=None, name: str | None = None, key=None, do_pickle: bool = False) -> CachedFunction:
    ...

class CachedFunction:
    def __init__(self, f: Callable, classmethod: bool = False,
                 name: str | None = None, key: Callable | None = None,
                 do_pickle: bool = False) -> None:
        ...

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        ...

    def cached(self, *args: Any, **kwds: Any) -> Any:
        ...

    def is_in_cache(self, *args: Any, **kwds: Any) -> bool:
        ...

    def set_cache(self, value: Any, *args: Any, **kwds: Any) -> None:
        ...

    def get_key(self, *args: Any, **kwds: Any) -> Any:
        ...

    def __repr__(self) -> str:
        ...

    def clear_cache(self) -> None:
        ...

    def precompute(self, arglist: Any, num_processes: int = 1) -> None:
        ...

class CachedMethod:
    def __init__(self, f: Callable, name: str | None = None,
                 key: Callable | None = None,
                 do_pickle: bool = False) -> None:
        ...

    def __call__(self, inst: Any, *args: Any, **kwds: Any) -> Any:
        ...

    def _get_instance_cache(self, inst: Any) -> dict:
        ...

    def __get__(self, inst: Any, cls: Any) -> Any:
        ...

class CacheDict(dict):
    pass

class CachedInParentMethod(CachedMethod):
    def __init__(self, f: Callable, name: str | None = None,
                 key: Callable | None = None,
                 do_pickle: bool = False) -> None:
        ...

    def _get_instance_cache(self, inst: Any) -> dict:
        ...

    def __get__(self, inst: Any, cls: Any) -> Any:
        ...

class CachedMethodCaller(CachedFunction):
    def __init__(self, cachedmethod: CachedMethod, inst: Any,
                 cache: dict | None = None, name: str | None = None,
                 key: Callable | None = None,
                 do_pickle: bool = False) -> None:
        ...

    def _instance_call(self, *args: Any, **kwds: Any) -> Any:
        ...

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        ...

    def cached(self, *args: Any, **kwds: Any) -> Any:
        ...

    def __get__(self, inst: Any, cls: Any) -> Any:
        ...

    def precompute(self, arglist: Any, num_processes: int = 1) -> None:
        ...

class CachedMethodCallerNoArgs(CachedFunction):
    def __init__(self, inst: Any, f: Callable, cache: Any = None,
                 name: str | None = None,
                 do_pickle: bool = False) -> None:
        ...

    def _instance_call(self) -> Any:
        ...

    def __call__(self) -> Any:
        ...

    def set_cache(self, value: Any) -> None:
        ...

    def clear_cache(self) -> None:
        ...

    def is_in_cache(self) -> bool:
        ...

    def __get__(self, inst: Any, cls: Any) -> Any:
        ...

class GloballyCachedMethodCaller(CachedMethodCaller):
    def get_key_args_kwds(self, args: tuple, kwds: dict) -> Any:
        ...



from typing import Any

def is_FpTElement(x: Any) -> bool:
    ...

class FpTElement:
    _numer: Any
    _denom: Any
    initialized: bool
    p: int

    def __init__(self, parent: Any, numer: Any, denom: Any = 1, coerce: bool = True, reduce: bool = True) -> None:
        ...

    def _new_c(self) -> 'FpTElement':
        ...

    def _copy_c(self) -> 'FpTElement':
        ...

    def numer(self) -> Any:
        ...

    def numerator(self) -> Any:
        ...

    def denom(self) -> Any:
        ...

    def denominator(self) -> Any:
        ...

    def __call__(self, *args: Any, **kwds: Any) -> 'FpTElement':
        ...

    def subs(self, in_dict: dict | None = None, *args: Any, **kwds: Any) -> 'FpTElement':
        ...

    def valuation(self, v: Any) -> int:
        ...

    def factor(self) -> Any:
        ...

    def _repr_(self) -> str:
        ...

    def _latex_(self) -> str:
        ...

    def _richcmp_(self, other: 'FpTElement', op: int) -> bool:
        ...

    def __hash__(self) -> int:
        ...

    def __neg__(self) -> 'FpTElement':
        ...

    def __invert__(self) -> 'FpTElement':
        ...

    def _add_(self, other: 'FpTElement') -> 'FpTElement':
        ...

    def _sub_(self, other: 'FpTElement') -> 'FpTElement':
        ...

    def _mul_(self, other: 'FpTElement') -> 'FpTElement':
        ...

    def _div_(self, other: 'FpTElement') -> 'FpTElement':
        ...

    def next(self) -> 'FpTElement':
        ...

    def _sqrt_or_None(self) -> 'FpTElement | None':
        ...

    def is_square(self) -> bool:
        ...

    def sqrt(self, extend: bool = True, all: bool = False) -> 'FpTElement | list[FpTElement]':
        ...

    def __pow__(self, e: int, dummy: Any) -> 'FpTElement':
        ...

class FpT:
    INTEGER_LIMIT: int

    def __init__(self, R: Any, names: Any = None) -> None:
        ...

    def __iter__(self) -> 'FpT_iter':
        ...

    def iter(self, bound: Any = None, start: Any = None) -> 'FpT_iter':
        ...

class FpT_iter:
    parent: Any
    degree: int
    cur: FpTElement
    g: Any

    def __init__(self, parent: Any, degree: int | None = None, start: FpTElement = None) -> None:
        ...

    def __cinit__(self, parent: Any, *args: Any, **kwds: Any) -> None:
        ...

    def __dealloc__(self) -> None:
        ...

    def __iter__(self) -> 'FpT_iter':
        ...

    def __next__(self) -> FpTElement:
        ...

class Polyring_FpT_coerce:
    p: int

    def __init__(self, R: Any) -> None:
        ...

    def _extra_slots(self) -> dict:
        ...

    def _update_slots(self, _slots: dict) -> None:
        ...

    def _call_(self, _x: Any) -> FpTElement:
        ...

    def _call_with_args(self, _x: Any, args: tuple = (), kwds: dict = {}) -> FpTElement:
        ...

    def section(self) -> 'FpT_Polyring_section':
        ...

class FpT_Polyring_section:
    p: int

    def __init__(self, f: Polyring_FpT_coerce) -> None:
        ...

    def _extra_slots(self) -> dict:
        ...

    def _update_slots(self, _slots: dict) -> None:
        ...

    def _call_(self, _x: Any) -> Any:
        ...

class Fp_FpT_coerce:
    p: int

    def __init__(self, R: Any) -> None:
        ...

    def _extra_slots(self) -> dict:
        ...

    def _update_slots(self, _slots: dict) -> None:
        ...

    def _call_(self, _x: Any) -> FpTElement:
        ...

    def _call_with_args(self, _x: Any, args: tuple = (), kwds: dict = {}) -> FpTElement:
        ...

    def section(self) -> 'FpT_Fp_section':
        ...

class FpT_Fp_section:
    p: int

    def __init__(self, f: Fp_FpT_coerce) -> None:
        ...

    def _extra_slots(self) -> dict:
        ...

    def _update_slots(self, _slots: dict) -> None:
        ...

    def _call_(self, _x: Any) -> Any:
        ...

class ZZ_FpT_coerce:
    p: int

    def __init__(self, R: Any) -> None:
        ...

    def _extra_slots(self) -> dict:
        ...

    def _update_slots(self, _slots: dict) -> None:
        ...

    def _call_(self, _x: Any) -> FpTElement:
        ...

    def _call_with_args(self, _x: Any, args: tuple = (), kwds: dict = {}) -> FpTElement:
        ...

    def section(self) -> Any:
        ...

def unpickle_FpT_element(K: Any, numer: Any, denom: Any) -> FpTElement:
    ...



# distutils: libraries = NTL_LIBRARIES gmp m
# distutils: extra_compile_args = NTL_CFLAGS
# distutils: include_dirs = NTL_INCDIR
# distutils: library_dirs = NTL_LIBDIR
# distutils: extra_link_args = NTL_LIBEXTRA
# distutils: language = c++

#*****************************************************************************
#       Copyright (C) 2005 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

include 'misc.pxi'
include 'decl.pxi'
import weakref

from sage.ext.cplusplus cimport ccrepr
from sage.rings.integer cimport Integer


cdef class ntl_ZZ_pContext_class():
    def __init__(self, ntl_ZZ v):
        """
        EXAMPLES::

            # You can construct contexts manually.
            sage: c = ntl.ZZ_pContext(11)
            sage: n1 = ntl.ZZ_p(12,c)
            sage: n1
            1

            # or You can construct contexts implicitly.
            sage: n2 = ntl.ZZ_p(12, 7)
            sage: n2
            5
            sage: ntl.ZZ_p(2,3)+ntl.ZZ_p(1,3)
            0
            sage: n2+n1  # Mismatched moduli:  It will go BOOM!
            Traceback (most recent call last):
            ...
            ValueError: You cannot perform arithmetic with elements of different moduli.
        """
        pass

    def __cinit__(self, ntl_ZZ v):
        self.x = ZZ_pContext_c(v.x)
        self.p = v
        self.p_bits = self.p._integer_().nbits()

    def __reduce__(self):
        """
        EXAMPLES::

            sage: c = ntl.ZZ_pContext(13)
            sage: loads(dumps(c)) is c
            True
        """
        return ntl_ZZ_pContext, (self.p,)

    def __repr__(self) -> str:
        """
        Return a print representation of ``self``.

        EXAMPLES::

            sage: c = ntl.ZZ_pContext(7)
            sage: c
            NTL modulus 7
        """
        return "NTL modulus %s" % (self.p)

    def __hash__(self):
        return hash(self.p)

    def modulus(self):
        """
        Return the current modulus associated to this
        context.

        EXAMPLES::

            sage: c = ntl.ZZ_pContext(7)
            sage: c.modulus()
            7

            sage: c = ntl.ZZ_pContext(10^30)
            sage: type(c.modulus())
            <class 'sage.rings.integer.Integer'>
            sage: c.modulus() == 10^30
            True
        """
        return Integer(self.p)

    def restore(self):
        """
        EXAMPLES::

            sage: c1 = ntl.ZZ_p(5,92) ; c2 = ntl.ZZ_p(7,92)
            sage: c1+c2
            12
            sage: d1 = ntl.ZZ_p(38,91) ; d2 = ntl.ZZ_p(3,91)
            sage: d1*d2 ## indirect doctest
            23
        """
        self.restore_c()

    cdef void restore_c(self) noexcept:
        self.x.restore()

    cpdef void _assert_is_current_modulus(self) except *:
        """
        Assert that is currently-set NTL modulus.

        Mostly for debugging purposes. If false, an assertion is raised. This method segfaults if
        the NTL modulus has never been set before.

        EXAMPLES::

            sage: c1 = ntl.ZZ_pContext(7)
            sage: c2 = ntl.ZZ_pContext(5)
            sage: c1.restore()
            sage: c1._assert_is_current_modulus()
            sage: c2._assert_is_current_modulus()
            Traceback (most recent call last):
            ...
            AssertionError: modulus mismatch: 5 != 7
            sage: c2.restore()
            sage: c1._assert_is_current_modulus()
            Traceback (most recent call last):
            ...
            AssertionError: modulus mismatch: 7 != 5
            sage: c2._assert_is_current_modulus()
            sage: ntl.ZZ_pContext(3).restore()
            sage: c1._assert_is_current_modulus()
            Traceback (most recent call last):
            ...
            AssertionError: modulus mismatch: 7 != 3
            sage: c2._assert_is_current_modulus()
            Traceback (most recent call last):
            ...
            AssertionError: modulus mismatch: 5 != 3
        """
        if self.p.x == ntl_ZZ_p_current_modulus():
            return
        raise AssertionError('modulus mismatch: {} != {}'.format(
            self.p,
            ccrepr(ntl_ZZ_p_current_modulus())))


cdef class ntl_ZZ_pContext_factory():

    def __init__(self):
        self.context_dict = {}

    cdef ntl_ZZ_pContext_class make_c(self, ntl_ZZ v):
        """
        Create a new ZZ_pContext.

        INPUT:

        - ``v`` -- an ntl_ZZ
        """
        cdef ntl_ZZ_pContext_class context
        if v in self.context_dict:
            context = <ntl_ZZ_pContext_class> self.context_dict[v]()
            if context is not None:
                return context
        context = ntl_ZZ_pContext_class(v)
        self.context_dict[v] = weakref.ref(context)
        return context


ZZ_pContext_factory = ntl_ZZ_pContext_factory()


def ntl_ZZ_pContext( v ):
    """
    Create a new ZZ_pContext.

    EXAMPLES::

        sage: c = ntl.ZZ_pContext(178)
        sage: n1 = ntl.ZZ_p(212,c)
        sage: n1
        34
    """
    v = ntl_ZZ(v)
    if (v < ntl_ZZ(2)):
        raise ValueError("%s is not a valid modulus." % v)
    return (<ntl_ZZ_pContext_factory>ZZ_pContext_factory).make_c(v)



# sage.doctest: needs sage.combinat sage.graphs
r"""
Graded rings of modular forms for Hecke triangle groups

AUTHORS:

- Jonas Jermann (2013): initial version
"""
# ****************************************************************************
#       Copyright (C) 2013-2014 Jonas Jermann <jjermann2@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.rings.integer_ring import ZZ
from sage.rings.infinity import infinity

from sage.structure.parent import Parent
from sage.categories.algebras import Algebras
from sage.structure.unique_representation import UniqueRepresentation

from .hecke_triangle_groups import HeckeTriangleGroup
from .abstract_ring import FormsRing_abstract


def canonical_parameters(group, base_ring, red_hom, n=None):
    r"""
    Return a canonical version of the parameters.

    EXAMPLES::

        sage: from sage.modular.modform_hecketriangle.graded_ring import canonical_parameters
        sage: canonical_parameters(4, ZZ, 1)
        (Hecke triangle group for n = 4, Integer Ring, True, 4)
        sage: canonical_parameters(infinity, RR, 0)
        (Hecke triangle group for n = +Infinity, Real Field with 53 bits of precision, False, +Infinity)
    """
    if n is not None:
        group = n

    if group == infinity:
        group = HeckeTriangleGroup(infinity)
    else:
        try:
            group = HeckeTriangleGroup(ZZ(group))
        except TypeError:
            group = HeckeTriangleGroup(group.n())

    red_hom = bool(red_hom)
    n = group.n()

    return (group, base_ring, red_hom, n)


class QuasiMeromorphicModularFormsRing(FormsRing_abstract, UniqueRepresentation):
    r"""
    Graded ring of (Hecke) quasi meromorphic modular forms
    for the given group and base ring.
    """

    @staticmethod
    def __classcall__(cls, group=HeckeTriangleGroup(3), base_ring=ZZ, red_hom=False, n=None):
        r"""
        Return a (cached) instance with canonical parameters.

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.graded_ring import (canonical_parameters, QuasiMeromorphicModularFormsRing)
            sage: (group, base_ring, red_hom, n) = canonical_parameters(4, ZZ, 1)
            sage: QuasiMeromorphicModularFormsRing(4, ZZ, 1) == QuasiMeromorphicModularFormsRing(group, base_ring, red_hom, n)
            True
        """
        (group, base_ring, red_hom, n) = canonical_parameters(group, base_ring, red_hom, n)
        return super().__classcall__(cls, group=group, base_ring=base_ring, red_hom=red_hom, n=n)

    def __init__(self, group, base_ring, red_hom, n):
        r"""
        Return the graded ring of (Hecke) quasi meromorphic modular forms
        for the given ``group`` and ``base_ring``.

        INPUT:

        - ``group`` -- the Hecke triangle group (default: ``HeckeTriangleGroup(3)``)

        - ``base_ring`` -- the base_ring (default: ``ZZ``)

        - ``red_hom`` -- if ``True`` then results of binary operations are
          considered homogeneous whenever it makes sense (default: ``False``).
          This is mainly used by the spaces of homogeneous elements.

        OUTPUT:

        The corresponding graded ring of (Hecke) quasi meromorphic modular forms
        for the given ``group`` and ``base_ring``.

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.graded_ring import QuasiMeromorphicModularFormsRing
            sage: MR = QuasiMeromorphicModularFormsRing(4, ZZ, 1)
            sage: MR
            QuasiMeromorphicModularFormsRing(n=4) over Integer Ring
            sage: MR.analytic_type()
            quasi meromorphic modular
            sage: MR.category()
            Category of commutative algebras over Integer Ring
            sage: MR in MR.category()
            True

            sage: QuasiMeromorphicModularFormsRing(n=infinity)
            QuasiMeromorphicModularFormsRing(n=+Infinity) over Integer Ring
        """
        FormsRing_abstract.__init__(self, group=group, base_ring=base_ring, red_hom=red_hom, n=n)
        cat = Algebras(base_ring).Commutative()
        Parent.__init__(self, category=cat)
        self._analytic_type = self.AT(["quasi", "mero"])


class QuasiWeakModularFormsRing(FormsRing_abstract, UniqueRepresentation):
    r"""
    Graded ring of (Hecke) quasi weakly holomorphic modular forms
    for the given group and base ring.
    """

    @staticmethod
    def __classcall__(cls, group=HeckeTriangleGroup(3), base_ring=ZZ, red_hom=False, n=None):
        r"""
        Return a (cached) instance with canonical parameters.

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.graded_ring import (canonical_parameters, QuasiWeakModularFormsRing)
            sage: (group, base_ring, red_hom, n) = canonical_parameters(5, CC, 0)
            sage: QuasiWeakModularFormsRing(5, CC, 0) == QuasiWeakModularFormsRing(group, base_ring, red_hom, n)
            True
        """
        (group, base_ring, red_hom, n) = canonical_parameters(group, base_ring, red_hom, n)
        return super().__classcall__(cls, group=group, base_ring=base_ring, red_hom=red_hom, n=n)

    def __init__(self, group, base_ring, red_hom, n):
        r"""
        Return the graded ring of (Hecke) quasi weakly holomorphic modular forms
        for the given ``group`` and ``base_ring``.

        INPUT:

        - ``group`` -- the Hecke triangle group (default: ``HeckeTriangleGroup(3)``)

        - ``base_ring`` -- the base_ring (default: ``ZZ``)

        - ``red_hom`` -- if ``True`` then results of binary operations are
          considered homogeneous whenever it makes sense (default: ``False``).
          This is mainly used by the spaces of homogeneous elements.

        OUTPUT:

        The corresponding graded ring of (Hecke) quasi weakly holomorphic modular forms
        for the given ``group`` and ``base_ring``.

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.graded_ring import QuasiWeakModularFormsRing
            sage: MR = QuasiWeakModularFormsRing(5, CC, 0)
            sage: MR
            QuasiWeakModularFormsRing(n=5) over Complex Field with 53 bits of precision
            sage: MR.analytic_type()
            quasi weakly holomorphic modular
            sage: MR.category()
            Category of commutative algebras over Complex Field with 53 bits of precision
            sage: MR in MR.category()
            True
        """
        FormsRing_abstract.__init__(self, group=group, base_ring=base_ring, red_hom=red_hom, n=n)
        cat = Algebras(base_ring).Commutative()
        Parent.__init__(self, category=cat)
        self._analytic_type = self.AT(["quasi", "weak"])


class QuasiModularFormsRing(FormsRing_abstract, UniqueRepresentation):
    r"""
    Graded ring of (Hecke) quasi modular forms
    for the given group and base ring
    """

    @staticmethod
    def __classcall__(cls, group=HeckeTriangleGroup(3), base_ring=ZZ, red_hom=False, n=None):
        r"""
        Return a (cached) instance with canonical parameters.

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.graded_ring import (canonical_parameters, QuasiModularFormsRing)
            sage: (group, base_ring, red_hom, n) = canonical_parameters(6, ZZ, True)
            sage: QuasiModularFormsRing(6, ZZ, True) == QuasiModularFormsRing(group, base_ring, red_hom, n)
            True
        """
        (group, base_ring, red_hom, n) = canonical_parameters(group, base_ring, red_hom, n)
        return super().__classcall__(cls, group=group, base_ring=base_ring, red_hom=red_hom, n=n)

    def __init__(self, group, base_ring, red_hom, n):
        r"""
        Return the graded ring of (Hecke) quasi modular forms
        for the given ``group`` and ``base_ring``.

        INPUT:

        - ``group`` -- the Hecke triangle group (default: ``HeckeTriangleGroup(3)``)

        - ``base_ring`` -- the base_ring (default: ``ZZ``)

        - ``red_hom`` -- if ``True`` then results of binary operations are
          considered homogeneous whenever it makes sense (default: ``False``).
          This is mainly used by the spaces of homogeneous elements.

        OUTPUT:

        The corresponding graded ring of (Hecke) quasi modular forms
        for the given ``group`` and ``base_ring``.

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.graded_ring import QuasiModularFormsRing
            sage: MR = QuasiModularFormsRing(6, ZZ, True)
            sage: MR
            QuasiModularFormsRing(n=6) over Integer Ring
            sage: MR.analytic_type()
            quasi modular
            sage: MR.category()
            Category of commutative algebras over Integer Ring
            sage: MR in MR.category()
            True
        """
        FormsRing_abstract.__init__(self, group=group, base_ring=base_ring, red_hom=red_hom, n=n)
        cat = Algebras(base_ring).Commutative()
        Parent.__init__(self, category=cat)
        self._analytic_type = self.AT(["quasi", "holo"])


class QuasiCuspFormsRing(FormsRing_abstract, UniqueRepresentation):
    r"""
    Graded ring of (Hecke) quasi cusp forms
    for the given group and base ring.
    """

    @staticmethod
    def __classcall__(cls, group=HeckeTriangleGroup(3), base_ring=ZZ, red_hom=False, n=None):
        r"""
        Return a (cached) instance with canonical parameters.

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.graded_ring import (canonical_parameters, QuasiCuspFormsRing)
            sage: (group, base_ring, red_hom, n) = canonical_parameters(7, ZZ, 1)
            sage: QuasiCuspFormsRing(7, ZZ, 1) == QuasiCuspFormsRing(group, base_ring, red_hom, n)
            True
        """
        (group, base_ring, red_hom, n) = canonical_parameters(group, base_ring, red_hom, n)
        return super().__classcall__(cls, group=group, base_ring=base_ring, red_hom=red_hom, n=n)

    def __init__(self, group, base_ring, red_hom, n):
        r"""
        Return the graded ring of (Hecke) quasi cusp forms
        for the given ``group`` and ``base_ring``.

        INPUT:

        - ``group`` -- the Hecke triangle group (default: ``HeckeTriangleGroup(3)``)

        - ``base_ring`` -- the base_ring (default: ``ZZ``)

        - ``red_hom`` -- if ``True`` then results of binary operations are
          considered homogeneous whenever it makes sense (default: ``False``).
          This is mainly used by the spaces of homogeneous elements.

        OUTPUT:

        The corresponding graded ring of (Hecke) quasi cusp forms
        for the given ``group`` and ``base_ring``.

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.graded_ring import QuasiCuspFormsRing
            sage: MR = QuasiCuspFormsRing(7, ZZ, 1)
            sage: MR
            QuasiCuspFormsRing(n=7) over Integer Ring
            sage: MR.analytic_type()
            quasi cuspidal
            sage: MR.category()
            Category of commutative algebras over Integer Ring
            sage: MR in MR.category()
            True
        """
        FormsRing_abstract.__init__(self, group=group, base_ring=base_ring, red_hom=red_hom, n=n)
        cat = Algebras(base_ring).Commutative()
        Parent.__init__(self, category=cat)
        self._analytic_type = self.AT(["quasi", "cusp"])


class MeromorphicModularFormsRing(FormsRing_abstract, UniqueRepresentation):
    r"""
    Graded ring of (Hecke) meromorphic modular forms
    for the given group and base ring
    """

    @staticmethod
    def __classcall__(cls, group=HeckeTriangleGroup(3), base_ring=ZZ, red_hom=False, n=None):
        r"""
        Return a (cached) instance with canonical parameters.

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.graded_ring import (canonical_parameters, MeromorphicModularFormsRing)
            sage: (group, base_ring, red_hom, n) = canonical_parameters(4, ZZ, 1)
            sage: MeromorphicModularFormsRing(4, ZZ, 1) == MeromorphicModularFormsRing(group, base_ring, red_hom, n)
            True
        """
        (group, base_ring, red_hom, n) = canonical_parameters(group, base_ring, red_hom, n)
        return super().__classcall__(cls, group=group, base_ring=base_ring, red_hom=red_hom, n=n)

    def __init__(self, group, base_ring, red_hom, n):
        r"""
        Return the graded ring of (Hecke) meromorphic modular forms
        for the given ``group`` and ``base_ring``.

        INPUT:

        - ``group`` -- the Hecke triangle group (default: ``HeckeTriangleGroup(3)``)

        - ``base_ring`` -- the base_ring (default: ``ZZ``)

        - ``red_hom`` -- if ``True`` then results of binary operations are
          considered homogeneous whenever it makes sense (default: ``False``).
          This is mainly used by the spaces of homogeneous elements.

        OUTPUT:

        The corresponding graded ring of (Hecke) meromorphic modular forms
        for the given ``group`` and ``base_ring``.

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.graded_ring import MeromorphicModularFormsRing
            sage: MR = MeromorphicModularFormsRing(4, ZZ, 1)
            sage: MR
            MeromorphicModularFormsRing(n=4) over Integer Ring
            sage: MR.analytic_type()
            meromorphic modular
            sage: MR.category()
            Category of commutative algebras over Integer Ring
            sage: MR in MR.category()
            True
        """
        FormsRing_abstract.__init__(self, group=group, base_ring=base_ring, red_hom=red_hom, n=n)
        cat = Algebras(base_ring).Commutative()
        Parent.__init__(self, category=cat)
        self._analytic_type = self.AT(["mero"])


class WeakModularFormsRing(FormsRing_abstract, UniqueRepresentation):
    r"""
    Graded ring of (Hecke) weakly holomorphic modular forms
    for the given group and base ring
    """

    @staticmethod
    def __classcall__(cls, group=HeckeTriangleGroup(3), base_ring=ZZ, red_hom=False, n=None):
        r"""
        Return a (cached) instance with canonical parameters.

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.graded_ring import (canonical_parameters, WeakModularFormsRing)
            sage: (group, base_ring, red_hom, n) = canonical_parameters(5, ZZ, 0)
            sage: WeakModularFormsRing(5, ZZ, 0) == WeakModularFormsRing(group, base_ring, red_hom, n)
            True
        """
        (group, base_ring, red_hom, n) = canonical_parameters(group, base_ring, red_hom, n)
        return super().__classcall__(cls, group=group, base_ring=base_ring, red_hom=red_hom, n=n)

    def __init__(self, group, base_ring, red_hom, n):
        r"""
        Return the graded ring of (Hecke) weakly holomorphic modular forms
        for the given ``group`` and ``base_ring``.

        INPUT:

        - ``group`` -- the Hecke triangle group (default: ``HeckeTriangleGroup(3)``)

        - ``base_ring`` -- the base_ring (default: ``ZZ``)

        - ``red_hom`` -- if ``True`` then results of binary operations are
          considered homogeneous whenever it makes sense (default: ``False``).
          This is mainly used by the spaces of homogeneous elements.

        OUTPUT:

        The corresponding graded ring of (Hecke) weakly holomorphic modular forms
        for the given ``group`` and ``base_ring``.

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.graded_ring import WeakModularFormsRing
            sage: MR = WeakModularFormsRing(5, ZZ, 0)
            sage: MR
            WeakModularFormsRing(n=5) over Integer Ring
            sage: MR.analytic_type()
            weakly holomorphic modular
            sage: MR.category()
            Category of commutative algebras over Integer Ring
            sage: MR in MR.category()
            True
        """
        FormsRing_abstract.__init__(self, group=group, base_ring=base_ring, red_hom=red_hom, n=n)
        cat = Algebras(base_ring).Commutative()
        Parent.__init__(self, category=cat)
        self._analytic_type = self.AT(["weak"])


class ModularFormsRing(FormsRing_abstract, UniqueRepresentation):
    r"""
    Graded ring of (Hecke) modular forms
    for the given group and base ring
    """

    @staticmethod
    def __classcall__(cls, group=HeckeTriangleGroup(3), base_ring=ZZ, red_hom=False, n=None):
        r"""
        Return a (cached) instance with canonical parameters.

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.graded_ring import ModularFormsRing
            sage: ModularFormsRing(3, ZZ, 0) == ModularFormsRing()
            True
        """
        (group, base_ring, red_hom, n) = canonical_parameters(group, base_ring, red_hom, n)
        return super().__classcall__(cls, group=group, base_ring=base_ring, red_hom=red_hom, n=n)

    def __init__(self, group, base_ring, red_hom, n):
        r"""
        Return the graded ring of (Hecke) modular forms
        for the given ``group`` and ``base_ring``.

        INPUT:

        - ``group`` -- the Hecke triangle group (default: ``HeckeTriangleGroup(3)``)

        - ``base_ring`` -- the base_ring (default: ``ZZ``)

        - ``red_hom`` -- if ``True`` then results of binary operations are
          considered homogeneous whenever it makes sense (default: ``False``).
          This is mainly used by the spaces of homogeneous elements.

        OUTPUT:

        The corresponding graded ring of (Hecke) modular forms
        for the given ``group`` and ``base_ring``.

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.graded_ring import ModularFormsRing
            sage: MR = ModularFormsRing()
            sage: MR
            ModularFormsRing(n=3) over Integer Ring
            sage: MR.analytic_type()
            modular
            sage: MR.category()
            Category of commutative algebras over Integer Ring
            sage: MR in MR.category()
            True
        """
        FormsRing_abstract.__init__(self, group=group, base_ring=base_ring, red_hom=red_hom, n=n)
        cat = Algebras(base_ring).Commutative()
        Parent.__init__(self, category=cat)
        self._analytic_type = self.AT(["holo"])


class CuspFormsRing(FormsRing_abstract, UniqueRepresentation):
    r"""
    Graded ring of (Hecke) cusp forms
    for the given group and base ring
    """

    @staticmethod
    def __classcall__(cls, group=HeckeTriangleGroup(3), base_ring=ZZ, red_hom=False, n=None):
        r"""
        Return a (cached) instance with canonical parameters.

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.graded_ring import (canonical_parameters, CuspFormsRing)
            sage: (group, base_ring, red_hom, n) = canonical_parameters(5, CC, True)
            sage: CuspFormsRing(5, CC, True) == CuspFormsRing(group, base_ring, red_hom, n)
            True
        """
        (group, base_ring, red_hom, n) = canonical_parameters(group, base_ring, red_hom, n)
        return super().__classcall__(cls, group=group, base_ring=base_ring, red_hom=red_hom, n=n)

    def __init__(self, group, base_ring, red_hom, n):
        r"""
        Return the graded ring of (Hecke) cusp forms
        for the given ``group`` and ``base_ring``.

        INPUT:

        - ``group`` -- the Hecke triangle group (default: ``HeckeTriangleGroup(3)``)

        - ``base_ring`` -- the base_ring (default: ``ZZ``)

        - ``red_hom`` -- if ``True`` then results of binary operations are
          considered homogeneous whenever it makes sense (default: ``False``).
          This is mainly used by the spaces of homogeneous elements.

        OUTPUT:

        The corresponding graded ring of (Hecke) cusp forms
        for the given ``group`` and ``base_ring``.

        EXAMPLES::

            sage: from sage.modular.modform_hecketriangle.graded_ring import CuspFormsRing
            sage: MR = CuspFormsRing(5, CC, True)
            sage: MR
            CuspFormsRing(n=5) over Complex Field with 53 bits of precision
            sage: MR.analytic_type()
            cuspidal
            sage: MR.category()
            Category of commutative algebras over Complex Field with 53 bits of precision
            sage: MR in MR.category()
            True

            sage: CuspFormsRing(n=infinity, base_ring=CC, red_hom=True)
            CuspFormsRing(n=+Infinity) over Complex Field with 53 bits of precision
        """
        FormsRing_abstract.__init__(self, group=group, base_ring=base_ring, red_hom=red_hom, n=n)
        cat = Algebras(base_ring).Commutative()
        Parent.__init__(self, category=cat)
        self._analytic_type = self.AT(["cusp"])



from sage.structure.list_clone cimport ClonableIntArray

cpdef list all_children(ClonableIntArray v, int max_part)
cpdef int lex_cmp_partial(ClonableIntArray t1, ClonableIntArray t2, int step) noexcept
cpdef int lex_cmp(ClonableIntArray t1, ClonableIntArray t2) noexcept
cpdef bint is_canonical(list sgs, ClonableIntArray v) except -1
cpdef ClonableIntArray canonical_representative_of_orbit_of(list sgs, ClonableIntArray v)
cpdef list canonical_children(list sgs, ClonableIntArray v, int max_part)
cpdef set orbit(list sgs, ClonableIntArray v)



from typing import Any

from sage.libs.gsl.types import gsl_complex
from sage.structure.element import FieldElement

class ComplexDoubleElement(FieldElement):
    _complex: gsl_complex

    def _new_c(self, x: gsl_complex) -> 'ComplexDoubleElement':
        ...

    def _add_(self, other: Any) -> Any:
        ...

    def _mul_(self, other: Any) -> Any:
        ...

    def _pow_(self, other: Any) -> Any:
        ...

def new_ComplexDoubleElement() -> ComplexDoubleElement:
    ...
